import inspect
import os
import shutil
import textwrap
from pathlib import Path

from conan import ConanFile
from conan.api.model import RecipeReference
from conan.api.output import ConanOutput
from conan.cli import make_abs_path
from conan.cli.printers.graph import print_graph_basic, print_graph_packages
from conan.errors import ConanException
from conan.internal.conan_app import ConanApp
from conan.internal.errors import conanfile_exception_formatter
from conan.internal.graph.install_graph import ProfileArgs
from conan.internal.methods import auto_language, auto_shared_fpic_config_options, \
    auto_shared_fpic_configure
from conan.internal.model.options import Options
from conan.internal.model.workspace import Workspace, WORKSPACE_YML, WORKSPACE_PY, WORKSPACE_FOLDER
from conan.tools.scm import Git
from conan.internal.graph.graph import (RECIPE_EDITABLE, DepsGraph, CONTEXT_HOST, RECIPE_VIRTUAL,
                                        Node, RECIPE_CONSUMER)
from conan.internal.graph.graph import TransitiveRequirement
from conan.internal.graph.profile_node_definer import consumer_definer, initialize_conanfile_profile
from conan.internal.loader import load_python_file
from conan.internal.source import retrieve_exports_sources
from conan.internal.util.files import merge_directories, save


def _find_ws_folder():
    path = Path(os.getcwd())
    while path.is_dir() and len(path.parts) > 1:  # finish at '/' or 'conanws/'
        if path.name == WORKSPACE_FOLDER:
            if (path / WORKSPACE_YML).is_file() or (path / WORKSPACE_PY).is_file():
                return str(path)
        if (path / WORKSPACE_YML).is_file() or (path / WORKSPACE_PY).is_file():
            return str(path)
        else:
            path = path.parent


def _load_workspace(ws_folder, conan_api):
    """ loads a conanfile basic object without evaluating anything, returns the module too
    """
    wspy = os.path.join(ws_folder, WORKSPACE_PY)
    if not os.path.isfile(wspy):
        ConanOutput().info(f"{WORKSPACE_PY} doesn't exist in {ws_folder}, using default behavior")
        assert os.path.exists(os.path.join(ws_folder, WORKSPACE_YML))
        ws = Workspace(ws_folder, conan_api)
    else:
        try:
            module, module_id = load_python_file(wspy)
            ws = _parse_module(module, module_id)
            ws = ws(ws_folder, conan_api)
        except ConanException as e:
            raise ConanException(f"Error loading {WORKSPACE_PY} at '{wspy}': {e}")
    return ws


def _parse_module(conanfile_module, module_id):
    result = None
    for name, attr in conanfile_module.__dict__.items():
        if (name.startswith("_") or not inspect.isclass(attr) or
                attr.__dict__.get("__module__") != module_id):
            continue

        if issubclass(attr, Workspace) and attr != Workspace:
            if result is None:
                result = attr
            else:
                raise ConanException("More than 1 Workspace in the file")

    if result is None:
        raise ConanException("No subclass of Workspace")

    return result


class WorkspaceAPI:
    TEST_ENABLED = False

    def __init__(self, conan_api):
        self._enabled = True
        self._conan_api = conan_api
        self._folder = _find_ws_folder()
        if self._folder:
            ConanOutput().warning(f"Workspace found: {self._folder}")
            if (WorkspaceAPI.TEST_ENABLED or os.getenv("CONAN_WORKSPACE_ENABLE")) != "will_break_next":
                ConanOutput().warning("Workspace ignored as CONAN_WORKSPACE_ENABLE is not set")
                self._folder = None
            else:
                ConanOutput().warning(f"Workspace is a dev-only feature, exclusively for testing")
                self._ws = _load_workspace(self._folder, conan_api)  # Error if not loading

    def enable(self, value):
        self._enabled = value

    def name(self):
        self._check_ws()
        return self._ws.name()

    def folder(self):
        """
        @return: the current workspace folder where the conanws.yml or conanws.py is located
        """
        self._check_ws()
        return self._folder

    def packages(self):
        """
        @return: Returns {RecipeReference: {"path": full abs-path, "output_folder": abs-path}}
        """
        if not self._folder or not self._enabled:
            return
        packages = {}
        for editable_info in self._ws.packages():
            rel_path = editable_info["path"]
            path = os.path.normpath(os.path.join(self._folder, rel_path, "conanfile.py"))
            if not os.path.isfile(path):
                raise ConanException(f"Workspace package not found: {path}")
            ref = editable_info.get("ref")
            try:
                if ref is None:
                    conanfile = self._ws.load_conanfile(rel_path)
                    reference = RecipeReference(name=conanfile.name, version=conanfile.version,
                                                user=conanfile.user, channel=conanfile.channel)
                else:
                    reference = RecipeReference.loads(ref)
                reference.validate_ref(reference)
            except Exception as e:
                raise ConanException(f"Workspace package reference could not be deduced by"
                                     f" {rel_path}/conanfile.py or it is not"
                                     f" correctly defined in the conanws.yml file: {e}")
            if reference in packages:
                raise ConanException(f"Workspace package '{str(reference)}' already exists.")
            packages[reference] = {"path": path}
            if editable_info.get("output_folder"):
                packages[reference]["output_folder"] = (
                    os.path.normpath(os.path.join(self._folder, editable_info["output_folder"]))
                )
        return packages

    def open(self, ref, remotes, cwd=None):
        cwd = cwd or os.getcwd()
        app = ConanApp(self._conan_api)
        ref = RecipeReference.loads(ref) if isinstance(ref, str) else ref
        recipe = app.proxy.get_recipe(ref, remotes, update=False, check_update=False)

        layout, recipe_status, remote = recipe
        if recipe_status == RECIPE_EDITABLE:
            raise ConanException(f"Can't open a dependency that is already an editable: {ref}")
        ref = layout.reference
        conanfile_path = layout.conanfile()
        conanfile, module = app.loader.load_basic_module(conanfile_path, remotes=remotes)

        scm = conanfile.conan_data.get("scm") if conanfile.conan_data else None
        dst_path = os.path.join(cwd, ref.name)
        if scm is None:
            conanfile.output.warning("conandata doesn't contain 'scm' information\n"
                                     "doing a local copy!!!")
            shutil.copytree(layout.export(), dst_path)
            retrieve_exports_sources(app.remote_manager, layout, conanfile, ref, remotes)
            export_sources = layout.export_sources()
            if os.path.exists(export_sources):
                conanfile.output.warning("There are export-sources, copying them, but the location"
                                         " might be incorrect, use 'scm' approach")
                merge_directories(export_sources, dst_path)
        else:
            git = Git(conanfile, folder=cwd)
            git.clone(url=scm["url"], target=ref.name)
            git.folder = ref.name  # change to the cloned folder
            git.checkout(commit=scm["commit"])
        return dst_path

    def _check_ws(self):
        if not self._folder:
            raise ConanException(f"Workspace not defined, please create a "
                                 f"'{WORKSPACE_PY}' or '{WORKSPACE_YML}' file")

    def add(self, path, name=None, version=None, user=None, channel=None, cwd=None,
            output_folder=None, remotes=None):
        """
        Add a new editable package to the current workspace (the current workspace must exist)
        @param path: The path to the folder containing the conanfile.py that defines the package
        @param name: (optional) The name of the package to be added if not defined in recipe
        @param version:
        @param user:
        @param channel:
        @param cwd:
        @param output_folder:
        @param remotes:
        @return: The reference of the added package
        """
        self._check_ws()
        full_path = self._conan_api.local.get_conanfile_path(path, cwd, py=True)
        app = ConanApp(self._conan_api)
        conanfile = app.loader.load_named(full_path, name, version, user, channel, remotes=remotes)
        if conanfile.name is None or conanfile.version is None:
            raise ConanException("Editable package recipe should declare its name and version")
        ref = RecipeReference(conanfile.name, conanfile.version, conanfile.user, conanfile.channel)
        ref.validate_ref()
        output_folder = make_abs_path(output_folder) if output_folder else None
        # Check the conanfile is there, and name/version matches
        self._ws.add(ref, full_path, output_folder)
        return ref

    def complete(self, profile_host, profile_build, lockfile, remotes, update):
        packages = self.packages()
        if not packages:
            ConanOutput().info("There are no packages in this workspace, nothing to complete")
            return

        for ref, info in packages.items():
            ConanOutput().title(f"Computing the dependency graph for package: {ref}")
            gapi = self._conan_api.graph
            deps_graph = gapi.load_graph_requires([ref], None, profile_host, profile_build,
                                                  lockfile, remotes, update)
            deps_graph.report_graph_error()
            print_graph_basic(deps_graph)

            nodes_to_complete = []
            for node in deps_graph.nodes[1:]:  # Exclude the current virtual root
                if node.recipe != RECIPE_EDITABLE:
                    # sanity check, a pacakge in the cache cannot have dependencies to the workspace
                    if any(d.node.recipe == RECIPE_EDITABLE for d in node.transitive_deps.values()):
                        nodes_to_complete.append(node)

            if not nodes_to_complete:
                ConanOutput().info("There are no intermediate packages to add to the workspace")
                return

            for node in nodes_to_complete:
                full_path = os.path.join(self._folder, node.name, "conanfile.py")
                dep_ref = node.ref
                ConanOutput().info(f"Adding to workspace {dep_ref}")
                try:
                    self._ws.add(dep_ref, full_path, output_folder=None)
                except ConanException:
                    if os.path.isfile(full_path):
                        raise
                    ConanOutput().info(f"Conanfile in {node.name} not found, trying "
                                       "to open it first")
                    self.open(dep_ref, remotes, cwd=self._folder)
                    self._ws.add(dep_ref, full_path, output_folder=None)

    @staticmethod
    def init(path):
        abs_path = make_abs_path(path)
        os.makedirs(abs_path, exist_ok=True)
        ws_yml_file = Path(abs_path, WORKSPACE_YML)
        ws_py_file = Path(abs_path, WORKSPACE_PY)
        if not ws_yml_file.exists():
            ConanOutput().success(f"Created empty {WORKSPACE_YML} in {path}")
            save(ws_yml_file, "")
        if not ws_py_file.exists():
            ConanOutput().success(f"Created minimal {WORKSPACE_PY} in {path}")
            ws_name = os.path.basename(abs_path)
            save(ws_py_file, textwrap.dedent(f'''\
            from conan import Workspace

            class MyWorkspace(Workspace):
               """
               Minimal Workspace class definition.
               More info: https://docs.conan.io/2/incubating.html#workspaces
               """
               def name(self):
                  return "{ws_name}"
            '''))

    def remove(self, path):
        self._check_ws()
        return self._ws.remove(path)

    def clean(self):
        self._check_ws()
        return self._ws.clean()

    def info(self):
        self._check_ws()
        return {"name": self._ws.name(),
                "folder": self._folder,
                "packages": self._ws.packages()}

    @staticmethod
    def _init_options(conanfile, options):
        if hasattr(conanfile, "config_options"):
            with conanfile_exception_formatter(conanfile, "config_options"):
                conanfile.config_options()
        elif "auto_shared_fpic" in conanfile.implements:
            auto_shared_fpic_config_options(conanfile)

        auto_language(conanfile)  # default implementation removes `compiler.cstd`

        # Assign only the current package options values, but none of the dependencies
        conanfile.options.apply_downstream(Options(), options, None, True)

        if hasattr(conanfile, "configure"):
            with conanfile_exception_formatter(conanfile, "configure"):
                conanfile.configure()
        elif "auto_shared_fpic" in conanfile.implements:
            auto_shared_fpic_configure(conanfile)

    def super_build_graph(self, deps_graph, profile_host, profile_build):
        order = []
        packages = self.packages()

        def find_folder(ref):
            return next(os.path.dirname(os.path.relpath(p["path"], self._folder)) for p_ref, p in
                        packages.items() if p_ref == ref)

        for level in deps_graph.by_levels():
            items = [item for item in level if item.recipe == "Editable"]
            level_order = []
            for node in items:
                conanfile = node.conanfile
                if hasattr(conanfile, "layout"):
                    with conanfile_exception_formatter(conanfile, "layout"):
                        conanfile.layout()
                base_folder = find_folder(node.ref)
                src_folder = os.path.normpath(os.path.join(base_folder, conanfile.folders.source))
                level_order.append({"ref": node.ref, "folder": src_folder.replace("\\", "/")})
            order.append(level_order)

        self._ws.build_order(order)

        ConanOutput().title("Collapsing workspace packages")

        root_class = self._ws.root_conanfile()
        if root_class is not None:
            conanfile = root_class(f"{WORKSPACE_PY} base project Conanfile")
            # To inject things like cmd_wrapper to the consumer conanfile, so self.run() works
            helpers = ConanApp(self._conan_api).loader._conanfile_helpers  # noqa
            conanfile._conan_helpers = helpers
            conanfile._conan_is_consumer = True
            initialize_conanfile_profile(conanfile, profile_build, profile_host, CONTEXT_HOST,
                                         is_build_require=False)
            # consumer_definer(conanfile, profile_host, profile_build)
            self._init_options(conanfile, profile_host.options)
            for field in ("requires", "build_requires", "test_requires", "requirements", "build",
                          "source", "package"):
                if getattr(conanfile, field, None):
                    raise ConanException(f"Conanfile in conanws.py shouldn't have '{field}'")
            root = Node(None, conanfile, context=CONTEXT_HOST, recipe=RECIPE_CONSUMER,
                        path=self._folder)  # path lets use the conanws.py folder
            root.should_build = True  # It is a consumer, this is something we are building
        else:
            ConanOutput().info(f"Workspace {WORKSPACE_PY} not found in the workspace folder, "
                               "using default behavior")
            conanfile = ConanFile(display_name="cli")
            consumer_definer(conanfile, profile_host, profile_build)
            root = Node(ref=None, conanfile=conanfile, context=CONTEXT_HOST, recipe=RECIPE_VIRTUAL)

        result = DepsGraph()  # TODO: We might need to copy more information from the original graph
        result.add_node(root)
        conanfile.workspace_packages = {}

        self._check_graph(deps_graph)
        for node in deps_graph.nodes[1:]:  # Exclude the current root
            if node.recipe != RECIPE_EDITABLE:
                result.add_node(node)
                continue
            # At the moment we are exposing the full conanfile, docs will warn against usage of
            # non pure functions
            conanfile.workspace_packages[node.ref] = node.conanfile
            for r, t in node.transitive_deps.items():
                if t.node.recipe == RECIPE_EDITABLE:
                    continue
                existing = root.transitive_deps.pop(r, None)
                if existing is None:
                    root.transitive_deps[r] = t
                else:
                    require = existing.require
                    require.aggregate(r)
                    root.transitive_deps[require] = TransitiveRequirement(require, t.node)

        # The graph edges must be defined too
        for r, t in root.transitive_deps.items():
            result.add_edge(root, t.node, r)

        return result

    @staticmethod
    def _check_graph(graph):
        for node in graph.nodes[1:]:  # Exclude the current root
            if node.recipe != RECIPE_EDITABLE:
                # sanity check, a pacakge in the cache cannot have dependencies to the workspace
                deps_edit = [d.node for d in node.transitive_deps.values()
                             if d.node.recipe == RECIPE_EDITABLE]
                if deps_edit:
                    raise ConanException(f"Workspace definition error. Package {node} in the "
                                         f"Conan cache has dependencies to packages "
                                         f"in the workspace: {deps_edit}\n"
                                         "Try the 'conan workspace complete' to open/add "
                                         "intermediate packages")

    def export(self, lockfile=None, remotes=None):
        self._check_ws()
        exported = []
        for ref, info in self.packages().items():
            exported_ref = self._conan_api.export.export(info["path"], ref.name, str(ref.version),
                                                         ref.user, ref.channel,
                                                         lockfile=lockfile, remotes=remotes)
            ref, _ = exported_ref
            exported.append(ref)
        return exported

    def select_packages(self, packages):
        self._check_ws()
        editable = self.packages()
        packages = packages or []
        selected_editables = {}
        for ref, info in editable.items():
            if packages and not any(ref.matches(p, False) for p in packages):
                continue
            selected_editables[ref] = info
        if not selected_editables:
            raise ConanException("There are no selected packages defined in the workspace")

        return selected_editables

    def build_order(self, packages, profile_host, profile_build, build_mode, lockfile, remotes,
                    profile_args, update=False):
        ConanOutput().title(f"Computing dependency graph for each package")
        conan_api = self._conan_api
        from conan.internal.graph.install_graph import InstallGraph
        install_order = InstallGraph(None)

        for ref, info in packages.items():
            ConanOutput().title(f"Computing the dependency graph for package: {ref}")

            deps_graph = conan_api.graph.load_graph_requires([ref], None,
                                                             profile_host, profile_build, lockfile,
                                                             remotes, update)
            deps_graph.report_graph_error()
            print_graph_basic(deps_graph)

            self._check_graph(deps_graph)

            conan_api.graph.analyze_binaries(deps_graph, build_mode, remotes=remotes, update=update,
                                             lockfile=lockfile)
            print_graph_packages(deps_graph)

            ConanOutput().success(f"\nAggregating build-order for package: {ref}")
            install_graph = InstallGraph(deps_graph, order_by="recipe",
                                         profile_args=ProfileArgs.from_args(profile_args))
            install_graph.raise_errors()
            install_order.merge(install_graph)

        return install_order



import re
import textwrap

import pytest

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient

new_value = "will_break_next"


@pytest.fixture
def client():
    c = TestClient()
    pkg = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.cmake import CMake, cmake_layout
        import os
        class Pkg(ConanFile):
            settings = "build_type", "os", "arch", "compiler"
            requires = "dep/0.1"
            generators = "CMakeDeps", "CMakeToolchain"
            def layout(self):  # Necessary to force config files in another location
                cmake_layout(self)
            def build(self):
                cmake = CMake(self)
                cmake.configure(variables={"CMAKE_FIND_DEBUG_MODE": "ON"})
       """)
    cmake = textwrap.dedent("""
       cmake_minimum_required(VERSION 3.15)
       project(pkgb LANGUAGES NONE)
       find_package(dep CONFIG REQUIRED)
       """)
    c.save({"dep/conanfile.py": GenConanfile("dep", "0.1"),
            "pkg/conanfile.py": pkg,
            "pkg/CMakeLists.txt": cmake})
    return c


@pytest.mark.tool("cmake")
def test_cmake_generated(client):
    c = client
    c.run("create dep")
    c.run(f"build pkg -c tools.cmake.cmakedeps:new={new_value}")
    assert "Conan toolchain: Including CMakeDeps generated conan_cmakedeps_paths.cmake" in c.out
    assert "Conan: Target declared imported INTERFACE library 'dep::dep'" in c.out


@pytest.mark.tool("cmake")
@pytest.mark.parametrize("lowercase", [False, True])
def test_cmake_in_package(client, lowercase):
    c = client
    # same, but in-package
    f = "dep-config" if lowercase else "depConfig"
    dep = textwrap.dedent(f"""
        import os
        from conan import ConanFile
        from conan.tools.files import save
        class Pkg(ConanFile):
            name = "dep"
            version = "0.1"

            def package(self):
                content = 'message(STATUS "Hello from dep dep-Config.cmake!!!!!")'
                save(self, os.path.join(self.package_folder, "cmake", "{f}.cmake"), content)
            def package_info(self):
                self.cpp_info.set_property("cmake_find_mode", "none")
                self.cpp_info.builddirs = ["cmake"]
        """)

    c.save({"dep/conanfile.py": dep})
    c.run("create dep")
    c.run(f"build pkg -c tools.cmake.cmakedeps:new={new_value}")
    assert "Conan toolchain: Including CMakeDeps generated conan_cmakedeps_paths.cmake" in c.out
    assert "Hello from dep dep-Config.cmake!!!!!" in c.out


class TestRuntimeDirs:

    def test_runtime_lib_dirs_multiconf(self):
        client = TestClient()
        app = GenConanfile().with_requires("dep/1.0").with_generator("CMakeDeps")\
            .with_settings("build_type")
        client.save({"lib/conanfile.py": GenConanfile(),
                     "dep/conanfile.py": GenConanfile("dep").with_requires("onelib/1.0",
                                                                           "twolib/1.0"),
                     "app/conanfile.py": app})
        client.run("create lib --name=onelib --version=1.0")
        client.run("create lib --name=twolib --version=1.0")
        client.run("create dep  --version=1.0")

        client.run(f'install app -s build_type=Release -c tools.cmake.cmakedeps:new={new_value}')
        client.run(f'install app -s build_type=Debug -c tools.cmake.cmakedeps:new={new_value}')

        contents = client.load("app/conan_cmakedeps_paths.cmake")
        pattern_lib_dirs = r"set\(CONAN_RUNTIME_LIB_DIRS ([^)]*)\)"
        runtime_lib_dirs = re.search(pattern_lib_dirs, contents).group(1)
        assert "<CONFIG:Release>" in runtime_lib_dirs
        assert "<CONFIG:Debug>" in runtime_lib_dirs
        # too simple of a check, but this is impossible to test automatically
        assert "set(CMAKE_VS_DEBUGGER_ENVIRONMENT" in contents


@pytest.mark.tool("cmake")
class TestCMakeDepsPaths:

    @pytest.mark.parametrize("requires, tool_requires", [(True, False), (False, True), (True, True)])
    def test_find_program_path(self, requires, tool_requires):
        """Test that executables in bindirs of tool_requires can be found with
        find_program() in consumer CMakeLists.
        """
        c = TestClient()

        conanfile = textwrap.dedent("""
            import os
            from conan.tools.files import copy
            from conan import ConanFile
            class TestConan(ConanFile):
                name = "tool"
                version = "1.0"
                exports_sources = "*"
                def package(self):
                    copy(self, "*", self.source_folder, os.path.join(self.package_folder, "bin"))
        """)
        c.save({"conanfile.py": conanfile, "hello": "", "hello.exe": ""})
        c.run("create .")

        requires = 'requires = "tool/1.0"' if requires else ""
        tool_requires = 'tool_requires = "tool/1.0"' if tool_requires else ""
        conanfile = textwrap.dedent(f"""
            from conan import ConanFile
            from conan.tools.cmake import CMake
            class PkgConan(ConanFile):
                {requires}
                {tool_requires}
                settings = "os", "arch", "compiler", "build_type"
                generators = "CMakeToolchain", "CMakeDeps"
                def build(self):
                    cmake = CMake(self)
                    cmake.configure()
        """)
        consumer = textwrap.dedent("""
            cmake_minimum_required(VERSION 3.15)
            project(MyHello NONE)
            find_program(HELLOPROG hello)
            if(HELLOPROG)
                message(STATUS "Found hello prog: ${HELLOPROG}")
            endif()
        """)
        c.save({"conanfile.py": conanfile, "CMakeLists.txt": consumer}, clean_first=True)
        c.run(f"build . -c tools.cmake.cmakedeps:new={new_value}")
        assert "Found hello prog" in c.out
        if requires and tool_requires:
            assert "There is already a 'tool/1.0' package contributing to CMAKE_PROGRAM_PATH" in c.out

    def test_find_include_and_lib_paths(self):
        c = TestClient()
        conanfile = textwrap.dedent("""
            import os
            from conan.tools.files import copy
            from conan import ConanFile
            class TestConan(ConanFile):
                name = "hello"
                version = "1.0"
                exports_sources = "*"
                def package(self):
                    copy(self, "*.h", self.source_folder, os.path.join(self.package_folder, "include"))
                    copy(self, "*.lib", self.source_folder, os.path.join(self.package_folder, "lib"))
                    copy(self, "*.a", self.source_folder, os.path.join(self.package_folder, "lib"))
                    copy(self, "*.so", self.source_folder, os.path.join(self.package_folder, "lib"))
                    copy(self, "*.dll", self.source_folder, os.path.join(self.package_folder, "lib"))
        """)
        c.save({"conanfile.py": conanfile,
                "hello.h": "", "hello.lib": "", "libhello.a": "",
                "libhello.so": "", "libhello.dll": ""})
        c.run("create .")
        conanfile = textwrap.dedent(f"""
            from conan import ConanFile
            from conan.tools.cmake import CMake
            class PkgConan(ConanFile):
                requires = "hello/1.0"
                settings = "os", "arch", "compiler", "build_type"
                generators = "CMakeToolchain", "CMakeDeps"
                def build(self):
                    cmake = CMake(self)
                    cmake.configure()
        """)
        consumer = textwrap.dedent("""
            cmake_minimum_required(VERSION 3.15)
            project(MyHello NONE)
            find_file(HELLOINC hello.h)
            find_library(HELLOLIB hello)
            if(HELLOINC)
                message(STATUS "Found hello header: ${HELLOINC}")
            endif()
            if(HELLOLIB)
                message(STATUS "Found hello lib: ${HELLOLIB}")
            endif()
        """)
        c.save({"conanfile.py": conanfile, "CMakeLists.txt": consumer}, clean_first=True)
        c.run(f"build . -c tools.cmake.cmakedeps:new={new_value}")
        assert "Found hello header" in c.out
        assert "Found hello lib" in c.out

    @pytest.mark.parametrize("require_type", ["requires", "tool_requires"])
    def test_include_modules(self, require_type):
        """Test that cmake module files in builddirs of requires and tool_requires
        are accessible with include() in consumer CMakeLists
        """
        c = TestClient()
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            from conan.tools.files import copy
            class TestConan(ConanFile):
                exports_sources = "*"
                def package(self):
                    copy(self, "*", self.source_folder, self.package_folder)
                def package_info(self):
                    self.cpp_info.builddirs.append("cmake")
        """)
        c.save({"conanfile.py": conanfile,
                "cmake/myowncmake.cmake": 'MESSAGE("MYOWNCMAKE FROM hello!")'})
        c.run("create . --name=hello --version=0.1")

        conanfile = textwrap.dedent(f"""
            from conan import ConanFile
            from conan.tools.cmake import CMake
            class PkgConan(ConanFile):
                settings = "os", "compiler", "arch", "build_type"
                {require_type} = "hello/0.1"
                generators = "CMakeToolchain", "CMakeConfigDeps"

                def build(self):
                    cmake = CMake(self)
                    cmake.configure()
                    cmake.build()
            """)
        consumer = textwrap.dedent("""
            cmake_minimum_required(VERSION 3.15)
            project(MyHello NONE)
            include(myowncmake)
        """)
        c.save({"conanfile.py": conanfile,
                "CMakeLists.txt": consumer}, clean_first=True)
        c.run(f"build . -c tools.cmake.cmakedeps:new={new_value}")
        assert "MYOWNCMAKE FROM hello!" in c.out

    def test_include_modules_both_build_host(self):
        c = TestClient()
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            from conan.tools.files import copy
            class TestConan(ConanFile):
                exports_sources = "*"
                def package(self):
                    copy(self, "*", self.source_folder, self.package_folder)
                def package_info(self):
                    self.cpp_info.builddirs.append("cmake")
            """)
        c.save({"conanfile.py": conanfile,
                "cmake/myowncmake.cmake": 'MESSAGE("MYOWNCMAKE FROM hello!")'})
        c.run("create . --name=hello --version=0.1")

        conanfile = textwrap.dedent(f"""
            from conan import ConanFile
            from conan.tools.cmake import CMake
            class PkgConan(ConanFile):
                settings = "os", "compiler", "arch", "build_type"
                requires = "hello/0.1"
                tool_requires = "hello/0.1"
                generators = "CMakeToolchain", "CMakeConfigDeps"

                def build(self):
                    cmake = CMake(self)
                    cmake.configure()
                    cmake.build()
            """)
        consumer = textwrap.dedent("""
            cmake_minimum_required(VERSION 3.15)
            project(MyHello NONE)
            include(myowncmake)
            """)
        c.save({"conanfile.py": conanfile,
                "CMakeLists.txt": consumer}, clean_first=True)
        c.run(f"build . -c tools.cmake.cmakedeps:new={new_value}")
        assert "conanfile.py: There is already a 'hello/0.1' package " \
               "contributing to CMAKE_MODULE_PATH" in c.out
        assert "MYOWNCMAKE FROM hello!" in c.out



import re
import textwrap

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient, NO_SETTINGS_PACKAGE_ID


def test_private_skip():
    # app -> pkg -(private)-> dep
    client = TestClient(light=True)
    client.save({"conanfile.py": GenConanfile()})
    client.run("create . --name=dep --version=1.0")
    client.save({"conanfile.py": GenConanfile().with_requirement("dep/1.0", visible=False)})
    client.run("create . --name=pkg --version=1.0")
    client.run("remove dep/1.0:* -c")  # Dep binary is removed not used at all

    client.save({"conanfile.py": GenConanfile().with_requires("pkg/1.0")})
    client.run("create . --name=app --version=1.0 -v")
    client.assert_listed_binary({"dep/1.0": (NO_SETTINGS_PACKAGE_ID, "Skip")})


def test_private_no_skip():
    # app -> pkg -(private)-> dep
    client = TestClient(light=True)
    client.save({"conanfile.py": GenConanfile()})
    client.run("create . --name=dep --version=1.0")
    client.save({"conanfile.py": GenConanfile().with_requirement("dep/1.0", visible=False)})
    client.run("create . --name=pkg --version=1.0")

    # But if we want to build pkg, no skip
    client.run("create . --name=app --version=1.0 --build=app/* --build=pkg/*")
    client.assert_listed_binary({"dep/1.0": (NO_SETTINGS_PACKAGE_ID, "Cache")})

    client.run("remove dep/1.0:* -c")  # Dep binary is removed not used at all
    client.run("create . --name=app --version=1.0 --build=app/* --build=pkg/*", assert_error=True)
    client.assert_listed_binary({"dep/1.0": (NO_SETTINGS_PACKAGE_ID, "Missing")})


def test_consumer_no_skip():
    # app -(private)-> pkg -> dep
    client = TestClient(light=True)
    client.save({"conanfile.py": GenConanfile()})
    client.run("create . --name=dep --version=1.0")
    client.save({"conanfile.py": GenConanfile().with_requires("dep/1.0")})
    client.run("create . --name=pkg --version=1.0")
    package_id = client.created_package_id("pkg/1.0")
    client.save({"conanfile.py": GenConanfile().with_requirement("pkg/1.0", visible=False)})

    client.run("install . ")

    client.assert_listed_binary({f"dep/1.0": (NO_SETTINGS_PACKAGE_ID, "Cache")})
    client.assert_listed_binary({f"pkg/1.0": (package_id, "Cache")})


def test_shared_link_static_skip():
    # app -> pkg (shared) -> dep (static)
    client = TestClient(light=True)
    client.save({"conanfile.py": GenConanfile().with_shared_option(False)})
    client.run("create . --name=dep --version=1.0")
    package_id = client.created_package_id("dep/1.0")
    client.save({"conanfile.py": GenConanfile().with_requirement("dep/1.0").
                with_shared_option(True)})
    client.run("create . --name=pkg --version=1.0")
    client.run("remove dep/1.0:* -c")  # Dep binary is removed not used at all

    client.save({"conanfile.py": GenConanfile().with_requires("pkg/1.0")})
    client.run("create . --name=app --version=1.0 -v")
    client.assert_listed_binary({"dep/1.0": (package_id, "Skip")})


def test_test_requires():
    # Using a test_requires can be skipped if it is not necessary to build its consumer
    # app -> pkg (static) -(test_requires)-> gtest (static)
    client = TestClient(light=True)
    client.save({"conanfile.py": GenConanfile().with_shared_option(False)})
    client.run("create . --name=gtest --version=1.0")
    package_id = client.created_package_id("gtest/1.0")
    client.save({"conanfile.py": GenConanfile().with_test_requires("gtest/1.0").
                with_shared_option(False)})
    client.run("create . --name=pkg --version=1.0")
    client.run("remove gtest/1.0:* -c")  # Dep binary is removed not used at all

    client.save({"conanfile.py": GenConanfile().with_requires("pkg/1.0")})
    # Checking list of skipped binaries
    client.run("create . --name=app --version=1.0")
    assert re.search(r"Skipped binaries(\s*)gtest/1.0", client.out)
    # Showing the complete information about the skipped binary
    client.run("create . --name=app --version=1.0 -v")
    client.assert_listed_binary({"gtest/1.0": (package_id, "Skip")}, test=True)


def test_build_scripts_no_skip():
    c = TestClient(light=True)
    c.save({"scripts/conanfile.py": GenConanfile("script", "0.1").with_package_type("build-scripts"),
            "app/conanfile.py": GenConanfile().with_tool_requires("script/0.1")})
    c.run("create scripts")
    c.assert_listed_binary({"script/0.1": ("da39a3ee5e6b4b0d3255bfef95601890afd80709", "Build")},
                           build=True)
    c.run("install app")
    c.assert_listed_binary({"script/0.1": ("da39a3ee5e6b4b0d3255bfef95601890afd80709", "Cache")},
                           build=True)


def test_list_skip_printing():
    """ make sure that when a package is required in the graph, it is not marked as SKIP, just
    because some other part of the graph is skipping it. In this case, a tool_require might be
    necessary for some packages building from soures, but not for others
    """
    c = TestClient(light=True)
    c.save({"tool/conanfile.py": GenConanfile("tool", "0.1"),
            "pkga/conanfile.py": GenConanfile("pkga", "0.1").with_tool_requires("tool/0.1"),
            "pkgb/conanfile.py": GenConanfile("pkgb", "0.1").with_requires("pkga/0.1")
                                                            .with_tool_requires("tool/0.1"),
            "app/conanfile.py": GenConanfile().with_requires("pkgb/0.1")})
    c.run("create tool")
    c.run("create pkga")
    c.run("create pkgb")
    c.run("remove pkga:* -c")
    c.run("install app --build=missing")
    c.assert_listed_binary({"tool/0.1": ("da39a3ee5e6b4b0d3255bfef95601890afd80709", "Cache")},
                           build=True)


def test_conf_skip():
    client = TestClient(light=True)
    client.save({"conanfile.py": GenConanfile()})
    client.run("create . --name=maths --version=1.0")
    client.run("create . --name=ai --version=1.0")

    client.save({"conanfile.py": GenConanfile().with_requirement("maths/1.0", visible=False)})
    client.run("create . --name=liba --version=1.0")
    client.save({"conanfile.py": GenConanfile().with_requirement("ai/1.0", visible=False)})
    client.run("create . --name=libb --version=1.0")

    client.save({"conanfile.py": GenConanfile().with_requires("liba/1.0", "libb/1.0")})
    client.run("create . --name=app --version=0.0 -v")
    client.assert_listed_binary({"maths/1.0": (NO_SETTINGS_PACKAGE_ID, "Skip")})
    client.assert_listed_binary({"ai/1.0": (NO_SETTINGS_PACKAGE_ID, "Skip")})

    client.run("create . --name=app --version=1.0 -v -c *:tools.graph:skip_binaries=False")
    client.assert_listed_binary({"maths/1.0": (NO_SETTINGS_PACKAGE_ID, "Cache")})
    client.assert_listed_binary({"ai/1.0": (NO_SETTINGS_PACKAGE_ID, "Cache")})

    client.run("create . --name=app --version=2.0 -v -c maths/*:tools.graph:skip_binaries=False")
    client.assert_listed_binary({"maths/1.0": (NO_SETTINGS_PACKAGE_ID, "Cache")})
    client.assert_listed_binary({"ai/1.0": (NO_SETTINGS_PACKAGE_ID, "Skip")})

    client.run("create . --name=app --version=3.0 -v -c *:tools.graph:skip_binaries=True")
    client.assert_listed_binary({"maths/1.0": (NO_SETTINGS_PACKAGE_ID, "Skip")})
    client.assert_listed_binary({"ai/1.0": (NO_SETTINGS_PACKAGE_ID, "Skip")})


def test_skipped_intermediate_header():
    # app -> libc/0.1 (static) -> libb0.1 (header) -> liba0.1 (static)
    # This libb0.1 cannot be skipped because it is necessary its lib-config.cmake for transitivity
    c = TestClient()
    c.save({"liba/conanfile.py": GenConanfile("liba", "0.1").with_package_type("static-library")
                                                            .with_package_info(cpp_info={"libs":
                                                                                         ["liba"]},
                                                                               env_info={}),
            "libb/conanfile.py": GenConanfile("libb", "0.1").with_package_type("header-library")
                                                            .with_requires("liba/0.1"),
            "libc/conanfile.py": GenConanfile("libc", "0.1").with_package_type("static-library")
                                                            .with_requires("libb/0.1"),
            "app/conanfile.py": GenConanfile("app", "0.1").with_requires("libc/0.1")
                                                          .with_settings("build_type")})
    c.run("create liba")
    c.run("create libb")
    c.run("create libc")
    c.run("install app -g CMakeDeps")
    c.assert_listed_binary({"liba/0.1": ("da39a3ee5e6b4b0d3255bfef95601890afd80709", "Cache"),
                            "libb/0.1": ("da39a3ee5e6b4b0d3255bfef95601890afd80709", "Cache"),
                            "libc/0.1": ("0f9f8919daed27aacd18c33199957e8882a87fd7", "Cache")})
    libc_data = c.load("app/libc-release-data.cmake")
    assert "list(APPEND libc_FIND_DEPENDENCY_NAMES libb)" in libc_data
    libb_data = c.load("app/libb-release-data.cmake")
    # libb brings no headers nor libraries
    assert "set(libb_INCLUDE_DIRS_RELEASE )" in libb_data
    assert "set(libb_LIBS_RELEASE )" in libb_data
    liba_data = c.load("app/liba-release-data.cmake")
    # liba brings only libraries
    assert "set(liba_INCLUDE_DIRS_RELEASE )" in liba_data
    assert "set(liba_LIBS_RELEASE liba)" in liba_data


def test_skip_visible_build():
    # https://github.com/conan-io/conan/issues/15346
    c = TestClient(light=True)
    c.save({"liba/conanfile.py": GenConanfile("liba", "0.1"),
            "libb/conanfile.py": GenConanfile("libb", "0.1").with_requirement("liba/0.1",
                                                                              build=True),
            "libc/conanfile.py": GenConanfile("libc", "0.1").with_requirement("libb/0.1",
                                                                              visible=False),
            "app/conanfile.py": GenConanfile("app", "0.1").with_requires("libc/0.1")})
    c.run("create liba")
    c.run("create libb")
    c.run("create libc")
    c.run("install app --format=json")
    assert re.search(r"Skipped binaries(\s*)libb/0.1, liba/0.1", c.out)


def test_skip_tool_requires_context():
    c = TestClient()
    cmake = textwrap.dedent("""
        from conan import ConanFile
        class CMake(ConanFile):
            name = "cmake"
            version = "1.0"
            def package_info(self):
                self.buildenv_info.define("MYVAR", "MYVALUE")
            """)
    gtest = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.files import load
        class gtest(ConanFile):
            name = "gtest"
            version = "1.0"
            settings = "os"
            package_type = "static-library"
            def build(self):
                env = load(self, "conanbuildenv.sh")
                self.output.info(f"MYENV: {env}")
        """)
    c.save({"cmake/conanfile.py": cmake,
            "gtest/conanfile.py": gtest,
            "lib/conanfile.py": GenConanfile("lib", "1.0").with_package_type("static-library")
                                                          .with_test_requires("gtest/1.0"),
            "app/conanfile.py": GenConanfile("app", "1.0").with_settings("os")
                                                          .with_requires("lib/1.0"),
            "profile": "[tool_requires]\ncmake/[>=1.0]"})

    c.run("create cmake")
    c.run("create gtest -s:a os=Linux")
    c.run("create lib -s:a os=Linux")

    c.run("remove gtest:* -c")
    c.run("install app -s:a os=Linux -pr=profile -c=tools.graph:skip_binaries=False --build=missing")
    assert 'export MYVAR="MYVALUE"' in c.out
    assert "gtest/1.0: Package '9a4eb3c8701508aa9458b1a73d0633783ecc2270' built" in c.out


def test_skip_intermediate_header():
    # https://github.com/conan-io/conan/issues/16402
    # Libb cannot be skipped in any case, because there is a link order libc->liba necessary
    # app -> libc/0.1 (static) -> libb0.1 (header) -> liba0.1 (static)
    #  \------------------------------------------------/
    # libb
    c = TestClient(light=True)
    c.save({"liba/conanfile.py": GenConanfile("liba", "0.1").with_package_type("static-library"),
            "libb/conanfile.py": GenConanfile("libb", "0.1").with_requirement("liba/0.1")
                                                            .with_package_type("header-library"),
            "libc/conanfile.py": GenConanfile("libc", "0.1").with_requirement("libb/0.1")
                                                            .with_package_type("static-library"),
            "app/conanfile.py": GenConanfile("app", "0.1").with_package_type("application")
                                                          .with_requires("libc/0.1", "liba/0.1")})
    c.run("create liba")
    c.run("create libb")
    c.run("create libc")
    c.run("install app")
    assert "Skipped binaries" not in c.out
    assert "libb/0.1: Already installed!" in c.out
    assert "liba/0.1: Already installed!" in c.out
    assert "libc/0.1: Already installed!" in c.out


def test_skip_intermediate_static():
    # https://github.com/conan-io/conan/issues/16402
    # In this case, libb can be completely skipped, because there is no linkage relationship at all
    # app -> libc/0.1 (shared) -> libb0.1 (static) -> liba0.1 (static)
    #  \------------------------------------------------/
    # libb
    c = TestClient(light=True)
    c.save({"liba/conanfile.py": GenConanfile("liba", "0.1").with_package_type("static-library"),
            "libb/conanfile.py": GenConanfile("libb", "0.1").with_requirement("liba/0.1")
                                                            .with_package_type("static-library"),
            "libc/conanfile.py": GenConanfile("libc", "0.1").with_requirement("libb/0.1")
                                                            .with_package_type("shared-library"),
            "app/conanfile.py": GenConanfile("app", "0.1").with_package_type("application")
                                                          .with_requires("libc/0.1", "liba/0.1")})
    c.run("create liba")
    c.run("create libb")
    c.run("create libc")
    c.run("remove libb:* -c")  # binary not necessary, can be skipped
    c.run("install app")
    assert re.search(r"Skipped binaries(\s*)libb/0.1", c.out)
    assert "libb/0.1: Already installed!" not in c.out
    assert "liba/0.1: Already installed!" in c.out
    assert "libc/0.1: Already installed!" in c.out


def test_skip_intermediate_static_complex():
    # https://github.com/conan-io/conan/issues/16402
    #  /----- libh(static)--libi(header)---libj(header)----------\
    # app -> libe(shared)->libd(static) -> libc(static) -> libb(static) -> liba(static)
    #  \---------libf(static) --libg(header)---------------------------------/
    # libd and libc can be skipped
    c = TestClient(light=True)
    c.save({"liba/conanfile.py": GenConanfile("liba", "0.1").with_package_type("static-library"),
            "libb/conanfile.py": GenConanfile("libb", "0.1").with_requirement("liba/0.1")
                                                            .with_package_type("static-library"),
            "libc/conanfile.py": GenConanfile("libc", "0.1").with_requirement("libb/0.1")
                                                            .with_package_type("static-library"),
            "libd/conanfile.py": GenConanfile("libd", "0.1").with_requirement("libc/0.1")
                                                            .with_package_type("static-library"),
            "libe/conanfile.py": GenConanfile("libe", "0.1").with_requirement("libd/0.1")
                                                            .with_package_type("shared-library"),
            "libg/conanfile.py": GenConanfile("libg", "0.1").with_requirement("liba/0.1")
                                                            .with_package_type("header-library"),
            "libf/conanfile.py": GenConanfile("libf", "0.1").with_requirement("libg/0.1")
                                                            .with_package_type("static-library"),
            "libj/conanfile.py": GenConanfile("libj", "0.1").with_requirement("libb/0.1")
                                                            .with_package_type("header-library"),
            "libi/conanfile.py": GenConanfile("libi", "0.1").with_requirement("libj/0.1")
                                                            .with_package_type("header-library"),
            "libh/conanfile.py": GenConanfile("libh", "0.1").with_requirement("libi/0.1")
                                                            .with_package_type("static-library"),
            "app/conanfile.py": GenConanfile("app", "0.1").with_package_type("application")
                                                          .with_requires("libh/0.1", "libe/0.1",
                                                                         "libf/0.1")
            })
    for lib in ("a", "b", "c", "d", "e", "j", "i", "h", "g", "f"):
        c.run(f"create lib{lib}")

    c.run("remove libd:* -c")  # binary not necessary, can be skipped
    c.run("remove libc:* -c")  # binary not necessary, can be skipped
    c.run("install app")
    assert re.search(r"Skipped binaries(\s*)libc/0.1, libd/0.1", c.out)
    assert "libd/0.1: Already installed!" not in c.out
    assert "libc/0.1: Already installed!" not in c.out
    for lib in ("a", "b", "e", "f", "g", "h", "i", "j"):
        assert f"lib{lib}/0.1: Already installed!" in c.out
    for lib in ("c", "d"):
        assert f"lib{lib}/0.1: Already installed!" not in c.out



import platform
import textwrap

import pytest

from conan.tools.cmake import CMakeToolchain
from conan.test.utils.tools import TestClient
from ._utils import create_library


@pytest.mark.skipif(platform.system() != "Darwin", reason="Requires XCode")
@pytest.mark.tool("cmake", "3.19")
def test_xcode_ios_generator():
    """ Simplest approach:
        https://cmake.org/cmake/help/latest/manual/cmake-toolchains.7.html#cross-compiling-for-ios-tvos-or-watchos
    """
    t = TestClient()
    create_library(t)
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.cmake import CMake, CMakeToolchain

        class Library(ConanFile):
            name = 'hello'
            version = '1.0'
            settings = 'os', 'arch', 'compiler', 'build_type'
            exports_sources = 'hello.h', 'hello.cpp', 'CMakeLists.txt'
            options = {'shared': [True, False]}
            default_options = {'shared': False}

            def generate(self):
                tc = CMakeToolchain(self, generator="Xcode")
                tc.generate()

            def build(self):
                cmake = CMake(self)
                cmake.configure()
                cmake.build()
                self.run("lipo -info Release-iphoneos/libhello.a")

            def package(self):
                cmake = CMake(self)
                cmake.install()
        """)

    ios_profile = textwrap.dedent("""
        [settings]
        os=iOS
        os.sdk=iphoneos
        os.version=12.0
        arch=armv8
        compiler=apple-clang
        compiler.version=12.0
        compiler.libcxx=libc++
        build_type=Release
        """)

    t.save({'conanfile.py': conanfile,
            "ios_profile": ios_profile})

    # Build in the cache
    t.run('create . --profile:build=default --profile:host=ios_profile')
    assert "Non-fat file: Release-iphoneos/libhello.a is architecture: arm64" in t.out

    # Build locally
    t.run('install . --profile:host=ios_profile --profile:build=default')
    t.run_command('cmake . -G"Xcode" -DCMAKE_TOOLCHAIN_FILE={}'.format(CMakeToolchain.filename))
    t.run_command('cmake --build . --config Release')
    t.run_command("lipo -info Release-iphoneos/libhello.a")
    assert "Non-fat file: Release-iphoneos/libhello.a is architecture: arm64" in t.out



import json
import os
import platform
import stat
import textwrap
from collections import OrderedDict

import pytest
from unittest.mock import patch
from requests import Response

from conan.errors import ConanException
from conan.api.model import PkgReference
from conan.internal.api.uploader import gzopen_without_timestamps
from conan.internal.paths import EXPORT_SOURCES_TGZ_NAME, PACKAGE_TGZ_NAME
from conan.test.utils.tools import NO_SETTINGS_PACKAGE_ID, TestClient, TestServer, \
    GenConanfile, TestRequester, TestingResponse
from conan.internal.util.files import is_dirty, save, set_dirty, sha1sum

conanfile = """from conan import ConanFile
from conan.tools.files import copy
class MyPkg(ConanFile):
    name = "hello0"
    version = "1.2.1"
    exports_sources = "*"

    def package(self):
        copy(self, "*.cpp", self.source_folder, self.package_folder)
        copy(self, "*.h", self.source_folder, self.package_folder)
"""


class TestUpload:

    @pytest.mark.artifactory_ready
    def test_upload_dirty(self):
        client = TestClient(default_server_user=True, light=True)
        client.save({"conanfile.py": GenConanfile("hello", "0.1")})
        client.run("create .")

        pkg_folder = client.created_layout().package()
        set_dirty(pkg_folder)

        client.run("upload * -r=default -c", assert_error=True)
        assert "ERROR: Package hello/0.1:da39a3ee5e6b4b0d3255bfef95601890afd80709 i" \
               "s corrupted, aborting upload." in client.out
        assert "Remove it with 'conan remove hello/0.1:da39a3ee5e6b4b0d3255bfef95601890afd80709" \
               in client.out

        # Test that removeing the binary allows moving forward
        client.run("remove hello/0.1:da39a3ee5e6b4b0d3255bfef95601890afd80709 -c")
        client.run("upload * -r=default --confirm")

    @pytest.mark.artifactory_ready
    def test_upload_force(self):
        client = TestClient(default_server_user=True, light=True)
        conanfile_ = textwrap.dedent("""
            from conan import ConanFile
            from conan.tools.files import copy
            class MyPkg(ConanFile):
                name = "hello"
                version = "0.1"
                def package(self):
                    copy(self, "myfile.sh", src=self.source_folder, dst=self.package_folder)
            """)
        client.save({"conanfile.py": conanfile_,
                    "myfile.sh": "foo"})

        client.run("export-pkg .")
        client.run("upload * --confirm -r default")
        assert "Uploading package 'hello" in client.out
        client.run("upload * --confirm -r default")
        assert "Uploading package" not in client.out

        if platform.system() == "Linux":
            package_file_path = os.path.join(client.current_folder, "myfile.sh")
            os.system('chmod +x "{}"'.format(package_file_path))
            assert os.stat(package_file_path).st_mode & stat.S_IXUSR
            client.run("export-pkg .")

            client.run("upload * --confirm -r default")
            # Doesn't change revision, doesn't reupload
            assert "conan_package.tgz" not in client.out
            assert "skipping upload" in client.out
            assert "Compressing package..." not in client.out

        # with --force it really re-uploads it
        client.run("upload * --confirm --force -r default")
        assert "Uploading recipe 'hello" in client.out
        assert "Uploading package 'hello" in client.out

        if platform.system() == "Linux":
            client.run("remove '*' -c")
            client.run("install --requires=hello/0.1 --deployer=full_deploy")
            package_file_path = os.path.join(client.current_folder, "full_deploy", "host", "hello",
                                             "0.1", "myfile.sh")
            # Owner with execute permissions
            assert os.stat(package_file_path).st_mode & stat.S_IXUSR

    @pytest.mark.artifactory_ready
    def test_pattern_upload(self):
        client = TestClient(default_server_user=True, light=True)
        client.save({"conanfile.py": conanfile})
        client.run("create . --user=user --channel=testing")
        client.run("upload hello0/*@user/testing --confirm -r default")
        assert "Uploading recipe 'hello0/1.2.1@" in client.out
        assert "Uploading package 'hello0/1.2.1@" in client.out

    def test_pattern_upload_no_recipes(self):
        client = TestClient(default_server_user=True, light=True)
        client.save({"conanfile.py": conanfile})
        client.run("upload bogus/*@dummy/testing --confirm -r default", assert_error=True)
        assert "No recipes found matching pattern 'bogus/*@dummy/testing'" in client.out

    def test_broken_sources_tgz(self):
        # https://github.com/conan-io/conan/issues/2854
        client = TestClient(default_server_user=True, light=True)
        client.save({"conanfile.py": conanfile,
                     "source.h": "my source"})
        client.run("create . --user=user --channel=testing")
        layout = client.exported_layout()

        def gzopen_patched(name, mode="r", fileobj=None, **kwargs):  # noqa
            raise ConanException("Error gzopen %s" % name)
        with patch('conan.internal.api.uploader.gzopen_without_timestamps', new=gzopen_patched):
            client.run("upload * --confirm -r default --only-recipe",
                       assert_error=True)
            assert "Error gzopen conan_sources.tgz" in client.out

            export_download_folder = layout.download_export()

            tgz = os.path.join(export_download_folder, EXPORT_SOURCES_TGZ_NAME)
            assert os.path.exists(tgz)
            assert is_dirty(tgz)

        client.run("upload * --confirm -r default --only-recipe")
        assert "Removing conan_sources.tgz, marked as dirty" in client.out
        assert os.path.exists(tgz)
        assert not is_dirty(tgz)

    def test_broken_package_tgz(self):
        # https://github.com/conan-io/conan/issues/2854
        client = TestClient(default_server_user=True, light=True)
        client.save({"conanfile.py": conanfile,
                     "source.h": "my source"})
        client.run("create . --user=user --channel=testing")
        pref = client.created_layout().reference

        def gzopen_patched(name, fileobj, compresslevel=None):  # noqa
            if name == PACKAGE_TGZ_NAME:
                raise ConanException("Error gzopen %s" % name)
            return gzopen_without_timestamps(name, fileobj)
        with patch('conan.internal.api.uploader.gzopen_without_timestamps', new=gzopen_patched):
            client.run("upload * --confirm -r default", assert_error=True)
            assert "Error gzopen conan_package.tgz" in client.out

            download_folder = client.get_latest_pkg_layout(pref).download_package()
            tgz = os.path.join(download_folder, PACKAGE_TGZ_NAME)
            assert os.path.exists(tgz)
            assert is_dirty(tgz)

        client.run("upload * --confirm -r default")
        assert "WARN: Removing conan_package.tgz, marked as dirty" in client.out
        assert os.path.exists(tgz)
        assert not is_dirty(tgz)

    def test_corrupt_upload(self):
        c = TestClient(default_server_user=True, light=True)

        c.save({"conanfile.py": conanfile,
                "include/hello.h": ""})
        c.run("create . --user=frodo --channel=stable")
        package_folder = c.created_layout().package()
        save(os.path.join(package_folder, "added.txt"), "")
        os.remove(os.path.join(package_folder, "include/hello.h"))
        c.run("upload hello0/1.2.1@frodo/stable --check -r default", assert_error=True)
        assert ("hello0/1.2.1@frodo/stable#3afd661184b94bdac7fb2057e7bd9baa"
                ":da39a3ee5e6b4b0d3255bfef95601890afd80709"
                "#e70e86439dec07a0d5d3414648b0b16c: ERROR") in c.out
        assert "include/hello.h (manifest: d41d8cd98f00b204e9800998ecf8427e, file: None)" in c.out
        assert "added.txt (manifest: None, file: d41d8cd98f00b204e9800998ecf8427e)" in c.out
        assert "ERROR: There are corrupted artifacts, check the error logs" in c.out

    @pytest.mark.artifactory_ready
    def test_upload_modified_recipe(self):
        client = TestClient(default_server_user=True, light=True)

        client.save({"conanfile.py": conanfile,
                     "hello.cpp": "int i=0"})
        client.run("export . --user=frodo --channel=stable")
        rrev = client.exported_recipe_revision()
        client.run("upload hello0/1.2.1@frodo/stable -r default")
        assert "Uploading recipe 'hello0/1.2.1@frodo/stable#" in client.out

        client2 = TestClient(servers=client.servers, inputs=["admin", "password"])
        client2.save({"conanfile.py": conanfile + "\r\n#end",
                      "hello.cpp": "int i=1"})
        client2.run("export . --user=frodo --channel=stable")
        layout = client2.exported_layout()
        manifest, _ = layout.recipe_manifests()
        manifest.time += 10
        manifest.save(layout.export())
        client2.run("upload hello0/1.2.1@frodo/stable -r default")
        assert "Uploading recipe 'hello0/1.2.1@frodo/stable#" in client2.out

        # first client tries to upload again
        # The client tries to upload exactly the same revision already uploaded, so no changes
        client.run("upload hello0/1.2.1@frodo/stable -r default")
        assert f"'hello0/1.2.1@frodo/stable#{rrev}' already in server, skipping upload" in client.out

    @pytest.mark.artifactory_ready
    def test_upload_unmodified_recipe(self):
        client = TestClient(default_server_user=True, light=True)
        files = {"conanfile.py": GenConanfile("hello0", "1.2.1")}
        client.save(files)
        client.run("export . --user=frodo --channel=stable")
        rrev = client.exported_recipe_revision()
        client.run("upload hello0/1.2.1@frodo/stable -r default")
        assert "Uploading recipe 'hello0/1.2.1@frodo/stable#" in client.out

        client2 = TestClient(servers=client.servers, inputs=["admin", "password"])
        client2.save(files)
        client2.run("export . --user=frodo --channel=stable")
        layout = client2.exported_layout()
        manifest, _ = layout.recipe_manifests()
        manifest.time += 10
        manifest.save(layout.export())
        client2.run("upload hello0/1.2.1@frodo/stable -r default")
        assert (f"Recipe 'hello0/1.2.1@frodo/stable#761f54e34d59deb172d6078add7050a7' already "
                f"in server, skipping upload") in client2.out

        # first client tries to upload again
        client.run("upload hello0/1.2.1@frodo/stable -r default")
        assert (f"Recipe 'hello0/1.2.1@frodo/stable#{rrev}' "
                f"already in server, skipping upload") in client.out

    @pytest.mark.artifactory_ready
    def test_upload_unmodified_package(self):
        client = TestClient(default_server_user=True, light=True)

        client.save({"conanfile.py": conanfile,
                     "hello.cpp": ""})
        client.run("create . --user=frodo --channel=stable")
        prev1 = client.created_layout().reference
        client.run("upload hello0/1.2.1@frodo/stable -r default")

        client2 = TestClient(servers=client.servers, inputs=["admin", "password"])
        client2.save({"conanfile.py": conanfile,
                      "hello.cpp": ""})
        client2.run("create . --user=frodo --channel=stable")
        prev2 = client2.created_layout().reference
        client2.run("upload hello0/1.2.1@frodo/stable -r default")
        assert f"'{repr(prev2.ref)}' already in server, skipping upload" in client2.out
        assert "Uploaded conan recipe 'hello0/1.2.1@frodo/stable' to 'default'" not in client2.out
        assert f"'{prev2.repr_notime()}' already in server, skipping upload" in client2.out

        # first client tries to upload again
        client.run("upload hello0/1.2.1@frodo/stable -r default")
        assert f"'{repr(prev1.ref)}' already in server, skipping upload" in client.out
        assert "Uploaded conan recipe 'hello0/1.2.1@frodo/stable' to 'default'" not in client.out
        assert f"'{prev1.repr_notime()}' already in server, skipping upload" in client2.out

    def test_upload_no_overwrite_all(self):
        conanfile_new = GenConanfile("hello", "1.0").\
            with_import("from conan.tools.files import copy").\
            with_exports_sources(["*"]).\
            with_package('copy(self, "*", self.source_folder, self.package_folder)')

        client = TestClient(default_server_user=True, light=True)
        client.save({"conanfile.py": conanfile_new,
                     "hello.h": ""})
        client.run("create . --user=frodo --channel=stable")
        prev1 = client.created_layout().reference
        # First time upload
        client.run("upload hello/1.0@frodo/stable -r default")
        assert "Forbidden overwrite" not in client.out
        assert "Uploading recipe 'hello/1.0@frodo/stable" in client.out

        # CASE: Upload again
        client.run("upload hello/1.0@frodo/stable -r default")
        assert f"'{repr(prev1.ref)}' already in server, skipping upload" in client.out
        assert f"'{prev1.repr_notime()}' already in server, skipping upload" in client.out

    def test_skip_upload(self):
        """ Check that the option --skip does not upload anything
        """
        client = TestClient(default_server_user=True, light=True)
        client.save({"conanfile.py": GenConanfile("hello0", "1.2.1").with_exports("*"),
                     "file.txt": ""})
        client.run("create .")

        client.run("upload * --dry-run -r default -c")
        assert "Compressing" in client.out
        client.run("search * -r default")
        # after dry run nothing should be on the server
        assert "hello" not in client.out

        # now upload, the stuff should NOT be recompressed
        client.run("upload * -c -r default")
        # check if compressed files are re-used
        assert "Compressing" not in client.out
        # now it should be on the server
        client.run("search * -r default")
        assert "hello0/1.2.1" in client.out

    def test_upload_without_sources(self):
        client = TestClient(default_server_user=True, light=True)
        client.save({"conanfile.py": GenConanfile()})
        client.run("create . --name=pkg --version=0.1 --user=user --channel=testing")
        client.run("upload * --confirm -r default")
        client2 = TestClient(servers=client.servers, inputs=["admin", "password",
                                                             "lasote", "mypass"])

        client2.run("install --requires=pkg/0.1@user/testing")
        client2.run("remote remove default")
        server2 = TestServer([("*/*@*/*", "*")], [("*/*@*/*", "*")],
                             users={"lasote": "mypass"})
        client2.servers = {"server2": server2}
        client2.update_servers()
        client2.run("upload * --confirm -r=server2")

        assert "Uploading recipe 'pkg" in client.out
        assert "Uploading package 'pkg" in client.out

    def test_upload_login_prompt_disabled_no_user(self):
        """ Without user info, uploads should fail when login prompt has been disabled.
        """
        files = {"conanfile.py": GenConanfile("hello0", "1.2.1")}
        client = TestClient(default_server_user=True, light=True)
        client.save(files)
        conan_conf = "core:non_interactive=True"
        client.save_home({"global.conf": conan_conf})

        client.run("create . --user=user --channel=testing")
        client.run("remote logout '*'")
        client.run("upload hello0/1.2.1@user/testing -r default", assert_error=True)

        assert "Conan interactive mode disabled" in client.out
        assert "-> conanmanifest.txt" not in client.out
        assert "-> conanfile.py" not in client.out
        assert "-> conan_export.tgz" not in client.out

    def test_upload_login_prompt_disabled_user_not_authenticated(self):
        # When a user is not authenticated, uploads should fail when login prompt has been disabled.
        files = {"conanfile.py": GenConanfile("hello0", "1.2.1")}
        client = TestClient(default_server_user=True, light=True)
        client.save(files)
        conan_conf = "core:non_interactive=True"
        client.save_home({"global.conf": conan_conf})
        client.run("create . --user=user --channel=testing")
        client.run("remote logout '*'")
        client.run("remote set-user default lasote")
        client.run("upload hello0/1.2.1@user/testing -r default", assert_error=True)
        assert "Conan interactive mode disabled" in client.out
        assert "-> conanmanifest.txt" not in client.out
        assert "-> conanfile.py" not in client.out
        assert "-> conan_export.tgz" not in client.out
        assert "Please enter a password for" not in client.out

    def test_upload_login_prompt_disabled_user_authenticated(self):
        #  When user is authenticated, uploads should work even when login prompt has been disabled.
        client = TestClient(default_server_user=True, light=True)
        client.save({"conanfile.py": GenConanfile("hello0", "1.2.1")})
        conan_conf = "core:non_interactive=True"
        client.save_home({"global.conf": conan_conf})
        client.run("create . --user=user --channel=testing")
        client.run("remote logout '*'")
        client.run("remote login default admin -p password")
        client.run("upload hello0/1.2.1@user/testing -r default")

        assert "Uploading recipe 'hello0/1.2.1@" in client.out
        assert "Uploading package 'hello0/1.2.1@" in client.out

    def test_upload_key_error(self):
        files = {"conanfile.py": GenConanfile("hello0", "1.2.1")}
        server1 = TestServer([("*/*@*/*", "*")], [("*/*@*/*", "*")], users={"lasote": "mypass"})
        server2 = TestServer([("*/*@*/*", "*")], [("*/*@*/*", "*")], users={"lasote": "mypass"})
        servers = OrderedDict()
        servers["server1"] = server1
        servers["server2"] = server2
        client = TestClient(servers=servers)
        client.save(files)
        client.run("create . --user=user --channel=testing")
        client.run("remote login server1 lasote -p mypass")
        client.run("remote login server2 lasote -p mypass")
        client.run("upload hello0/1.2.1@user/testing -r server1")
        client.run("remove * --confirm")
        client.run("install --requires=hello0/1.2.1@user/testing -r server1")
        client.run("remote remove server1")
        client.run("upload hello0/1.2.1@user/testing -r server2")
        assert "ERROR: 'server1'" not in client.out

    def test_upload_without_user_channel(self):
        server = TestServer(users={"user": "password"}, write_permissions=[("*/*@*/*", "*")])
        servers = {"default": server}
        client = TestClient(servers=servers, inputs=["user", "password"])

        client.save({"conanfile.py": GenConanfile()})

        client.run('create . --name=lib --version=1.0')
        assert "lib/1.0: Package '{}' created".format(NO_SETTINGS_PACKAGE_ID) in client.out
        client.run('upload lib/1.0 -c -r default')
        assert "Uploading recipe 'lib/1.0" in client.out

        # Verify that in the remote it is stored as "_"
        pref = PkgReference.loads("lib/1.0@#0:{}#0".format(NO_SETTINGS_PACKAGE_ID))
        path = server.server_store.export(pref.ref)
        assert "/lib/1.0/_/_/0/export" in path.replace("\\", "/")

        path = server.server_store.package(pref)
        assert "/lib/1.0/_/_/0/package" in path.replace("\\", "/")

        # Should be possible with explicit package
        client.run(f'upload lib/1.0#*:{NO_SETTINGS_PACKAGE_ID} -c -r default --force')
        assert "Uploading artifacts" in client.out

    def test_upload_without_cleaned_user(self):
        """ When a user is not authenticated, uploads failed first time
        https://github.com/conan-io/conan/issues/5878
        """

        class EmptyCapabilitiesResponse(object):
            def __init__(self):
                self.ok = False
                self.headers = {"X-Conan-Server-Capabilities": "",
                                "Content-Type": "application/json"}
                self.status_code = 401
                self.content = b''

        class ServerCapabilitiesRequester(TestRequester):
            def __init__(self, *args, **kwargs):
                self._first_ping = True
                super(ServerCapabilitiesRequester, self).__init__(*args, **kwargs)

            def get(self, url, **kwargs):
                app, url = self._prepare_call(url, kwargs)
                assert app
                assert ("/v1/" in url and url.endswith("ping")) or "/v2" in url
                if url.endswith("ping") and self._first_ping:
                    self._first_ping = False
                    return EmptyCapabilitiesResponse()
                else:
                    response = app.get(url, **kwargs)
                    return TestingResponse(response)

        server = TestServer(users={"user": "password"}, write_permissions=[("*/*@*/*", "*")])
        servers = {"default": server}
        client = TestClient(requester_class=ServerCapabilitiesRequester, servers=servers,
                            inputs=["user", "password"])
        files = {"conanfile.py": GenConanfile("hello0", "1.2.1")}
        client.save(files)
        client.run("create . --user=user --channel=testing")
        client.run("remote logout '*'")
        client.run("upload hello0/1.2.1@user/testing -r default")
        assert "Uploading recipe 'hello0/1.2.1@user/testing" in client.out

    def test_server_returns_200_ok(self):
        # https://github.com/conan-io/conan/issues/16104
        # If server returns 200 ok, without headers, it raises an error
        class MyHttpRequester(TestRequester):
            def get(self, _, **kwargs):
                resp = Response()
                resp.status_code = 200
                return resp

        client = TestClient(requester_class=MyHttpRequester, servers={"default": TestServer()})
        client.save({"conanfile.py": GenConanfile("hello0", "1.2.1")})
        client.run("create . ")
        client.run("upload * -c -r default", assert_error=True)
        assert "doesn't seem like a valid Conan remote" in client.out


def test_upload_only_without_user_channel():
    """
    check that we can upload only the packages without user and channel
    https://github.com/conan-io/conan/issues/10579
    """
    c = TestClient(default_server_user=True, light=True)

    c.save({"conanfile.py": GenConanfile("lib", "1.0")})
    c.run('create .')
    c.run("create . --user=user --channel=channel")
    c.run("list *")
    assert "lib/1.0@user/channel" in c.out

    c.run('upload */*@ -c -r=default')
    assert "Uploading recipe 'lib/1.0" in c.out  # FAILS!
    assert "lib/1.0@user/channel" not in c.out
    c.run("search * -r=default")
    assert "lib/1.0" in c.out
    assert "lib/1.0@user/channel" not in c.out

    c.run('upload */*@user/channel -c -r=default')
    assert "Uploading recipe 'lib/1.0@user/channel" in c.out
    c.run("search * -r=default")
    assert "lib/1.0@user/channel" in c.out
    assert "lib/1.0" in c.out


def test_upload_with_python_requires():
    # https://github.com/conan-io/conan/issues/14503
    c = TestClient(default_server_user=True, light=True)
    c.save({"tool/conanfile.py": GenConanfile("tool", "0.1"),
            "dep/conanfile.py": GenConanfile("dep", "0.1").with_python_requires("tool/[>=0.1]")})
    c.run("create tool")
    c.run("create dep")
    c.run("upload tool* -c -r=default")
    c.run("remove tool* -c")
    c.run("upload dep* -c -r=default")
    # This used to fail, but adding the enabled remotes to python_requires resolution, it works
    assert "tool/0.1: Downloaded recipe" in c.out


def test_upload_list_only_recipe():
    c = TestClient(default_server_user=True, light=True)
    c.save({"conanfile.py": GenConanfile("liba", "0.1")})
    c.run("create .")
    c.run("install --requires=liba/0.1 --format=json", redirect_stdout="graph.json")
    c.run("list --graph=graph.json --format=json", redirect_stdout="installed.json")
    c.run("upload --list=installed.json --only-recipe -r=default -c")
    assert "conan_package.tgz" not in c.out


@pytest.mark.parametrize("dry_run", [True, False])
def test_upload_json_output(dry_run):
    c = TestClient(default_server_user=True, light=True)
    c.save({"conanfile.py": GenConanfile("liba", "0.1").with_settings("os")
                                                       .with_shared_option(False)})
    c.run("create . -s os=Linux")
    dry_run_arg = "--dry-run" if dry_run else ""
    c.run(f"upload * -r=default {dry_run_arg} -c --format=json")
    list_pkgs = json.loads(c.stdout)
    revs = list_pkgs["default"]["liba/0.1"]["revisions"]["a565bd5defd3a99e157698fcc6e23b25"]
    pkg = revs["packages"]["9e0f8140f0fe6b967392f8d5da9881e232e05ff8"]
    prev = pkg["revisions"]["f50f552c6e04b1f241e5f7864bc3957f"]
    assert pkg["info"] == {"settings": {"os": "Linux"}, "options": {"shared": "False"}}
    base_url = "v2/conans/liba/0.1/_/_/revisions/a565bd5defd3a99e157698fcc6e23b25"
    assert revs["upload-urls"] == {
        "conanfile.py": {
            "url": f"{c.servers['default']}/{base_url}/files/conanfile.py",
            "checksum": sha1sum(revs["files"]["conanfile.py"])
        },
        "conanmanifest.txt": {
            "url": f"{c.servers['default']}/{base_url}/files/conanmanifest.txt",
            "checksum": sha1sum(revs["files"]["conanmanifest.txt"])
        }
    }
    pkg_url = "9e0f8140f0fe6b967392f8d5da9881e232e05ff8/revisions/f50f552c6e04b1f241e5f7864bc3957f"
    assert prev["upload-urls"] == {
        "conan_package.tgz": {
            "url": f"{c.servers['default']}/{base_url}/packages/{pkg_url}/files/conan_package.tgz",
            "checksum": sha1sum(prev["files"]["conan_package.tgz"])
        },
        "conaninfo.txt": {
            "url": f"{c.servers['default']}/{base_url}/packages/{pkg_url}/files/conaninfo.txt",
            "checksum": sha1sum(prev["files"]["conaninfo.txt"])
        },
        "conanmanifest.txt": {
            "url": f"{c.servers['default']}/{base_url}/packages/{pkg_url}/files/conanmanifest.txt",
            "checksum": sha1sum(prev["files"]["conanmanifest.txt"])
        }
    }

    if dry_run:
        # check we don't have anything about the upload-urls in the text formatter
        c.run("upload * -r=default -c --dry-run")
        assert "upload-urls" not in c.out
        assert "url:" not in c.out
        assert "checksum:" not in c.out



import requests
from requests.models import Response

from conan.test.utils.tools import TestClient

resp = Response()
resp._content = b'{"results": []}'
resp.status_code = 200
resp.headers = {"Content-Type": "application/json", "X-Conan-Server-Capabilities": "revisions"}


class RequesterMockFalse:
    expected = False

    def __init__(self, *args, **kwargs):
        pass

    def get(self, url, *args, **kwargs):  # noqa
        assert kwargs["verify"] is self.expected
        return resp

    def Session(self):  # noqa
        return self

    @property
    def codes(self):
        return requests.codes

    def mount(self, *args, **kwargs):
        pass


class RequesterMockTrue(RequesterMockFalse):
    expected = True


class TestVerifySSL:

    def test_verify_ssl(self):
        c = TestClient(requester_class=RequesterMockTrue)
        c.run("remote add myremote https://localhost --insecure")
        c.run("remote list")
        assert "Verify SSL: False" in c.out

        c.run("remote update myremote --url https://localhost --secure")
        c.run("remote list")
        assert "Verify SSL: True" in c.out

        c.run("remote remove myremote")
        c.run("remote add myremote https://localhost")
        c.run("remote list")
        assert "Verify SSL: True" in c.out

        # Verify that SSL is checked in requests
        c.run("search op* -r myremote")
        assert "WARN: There are no matching recipe references" in c.out

        # Verify that SSL is not checked in requests
        c = TestClient(requester_class=RequesterMockFalse)
        c.run("remote add myremote https://localhost --insecure")
        c.run("search op* -r myremote")
        assert "WARN: There are no matching recipe references" in c.out



import textwrap

import pytest

from conan.test.utils.tools import TestClient, GenConanfile


class TestOptions:

    def test_general_scope_options_test_package(self):
        client = TestClient()
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            class Pkg(ConanFile):
                options = {"shared": ["1", "2"]}
                def configure(self):
                    self.output.info("BUILD SHARED: %s" % self.options.shared)
            """)
        test = GenConanfile().with_test("pass")
        client.save({"conanfile.py": conanfile})
        client.run("create . --name=pkg --version=0.1 --user=user --channel=testing -o *:shared=1")
        assert "pkg/0.1@user/testing: BUILD SHARED: 1" in client.out
        client.run("create . --name=pkg --version=0.1 --user=user --channel=testing -o shared=2")
        assert 'legacy: Unscoped option definition is ambiguous' in client.out
        assert "pkg/0.1@user/testing: BUILD SHARED: 2" in client.out
        # With test_package
        client.save({"conanfile.py": conanfile,
                     "test_package/conanfile.py": test})
        client.run("create . --name=pkg --version=0.1 --user=user --channel=testing -o *:shared=1")
        assert "pkg/0.1@user/testing: BUILD SHARED: 1" in client.out
        client.run("create . --name=pkg --version=0.1 --user=user --channel=testing -o pkg*:shared=2")
        assert "pkg/0.1@user/testing: BUILD SHARED: 2" in client.out
        client.run("create . --name=pkg --version=0.1 --user=user --channel=testing -o shared=1")
        assert "pkg/0.1@user/testing: BUILD SHARED: 1" in client.out

    def test_general_scope_options_test_package_notdefined(self):
        client = TestClient()
        conanfile = GenConanfile()
        client.save({"conanfile.py": conanfile})
        client.run("create . --name=pkg --version=0.1 --user=user --channel=testing -o *:shared=True")
        assert "pkg/0.1@user/testing: Forced build from source" in client.out
        client.run("create . --name=pkg --version=0.1 --user=user --channel=testing -o shared=False", assert_error=True)
        assert "option 'shared' doesn't exist" in client.out
        # With test_package
        client.save({"conanfile.py": conanfile,
                     "test_package/conanfile.py": GenConanfile().with_test("pass")})
        client.run("create . --name=pkg --version=0.1 --user=user --channel=testing -o *:shared=True")
        assert "pkg/0.1@user/testing: Forced build from source" in client.out
        assert "Testing the package: Building" in client.out

    def test_general_scope_priorities(self):
        client = TestClient()
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            class Pkg(ConanFile):
                options = {"shared": ["1", "2", "3"], "other": [4, 5, 6]}
                def configure(self):
                    self.output.info("BUILD SHARED: %s OTHER: %s"
                                     % (self.options.shared, self.options.other))
            """)
        client.save({"conanfile.py": conanfile})
        # Consumer has priority
        client.run("create . --name=pkg --version=0.1 -o *:shared=1 -o shared=2 -o p*:other=4")
        assert "pkg/0.1: BUILD SHARED: 2 OTHER: 4" in client.out
        # Consumer has priority over pattern, even if the pattern specifies the package name
        client.run("create . --name=pkg --version=0.1 -o *:shared=1 -o pkg/*:shared=2 -o shared=3 -o p*:other=4")
        assert "pkg/0.1: BUILD SHARED: 3 OTHER: 4" in client.out
        client.run("create . --name=pkg --version=0.1 -o pkg/0.1:shared=2 -o p*:other=4 -o pk*:other=5")
        assert "pkg/0.1: BUILD SHARED: 2 OTHER: 5" in client.out

        # With test_package
        client.save({"conanfile.py": conanfile,
                     "test_package/conanfile.py": GenConanfile().with_test("pass")})
        # Sorted (longest, alphabetical) patterns, have priority
        client.run("create . --name=pkg --version=0.1 -o *:shared=1 -o pkg/0.1:shared=2 -o other=4")
        assert "pkg/0.1: BUILD SHARED: 2 OTHER: 4" in client.out
        client.run("create . --name=pkg --version=0.1 -o pk*:shared=2 -o p*:shared=1 -o pkg/0.1:other=5")
        assert "pkg/0.1: BUILD SHARED: 1 OTHER: 5" in client.out
        client.run("create . --name=pkg --version=0.1 -o pk*:shared=2 -o p*:shared=1 -o pkg/0.1:other=5 -o *g*:other=6")
        assert "pkg/0.1: BUILD SHARED: 1 OTHER: 6" in client.out

    def test_parsing(self):
        client = TestClient()
        conanfile = '''
from conan import ConanFile
class EqualerrorConan(ConanFile):
    name = "equal"
    version = "1.0.0"
    options = {"opt": ["ANY"]}
    default_options = {"opt": "b=c"}

    def build(self):
        self.output.warning("OPTION %s" % self.options.opt)
'''
        client.save({"conanfile.py": conanfile})
        client.run("export . --user=user --channel=testing")
        conanfile = '''
[requires]
equal/1.0.0@user/testing
[options]
equal/1.0.0@user/testing:opt=a=b
'''
        client.save({"conanfile.txt": conanfile}, clean_first=True)
        client.run("install . --build=missing")
        assert "OPTION a=b" in client.out

    def test_general_scope_options(self):
        # https://github.com/conan-io/conan/issues/2538
        client = TestClient()
        conanfile_liba = textwrap.dedent("""
            from conan import ConanFile
            class LibA(ConanFile):
                options = {"shared": [True, False]}

                def configure(self):
                    self.output.info("shared=%s" % self.options.shared)
                """)
        client.save({"conanfile.py": conanfile_liba})
        client.run("create . --name=liba --version=0.1 --user=danimtb --channel=testing -o *:shared=True")
        assert "liba/0.1@danimtb/testing: shared=True" in client.out

        conanfile_libb = textwrap.dedent("""
            from conan import ConanFile
            class LibB(ConanFile):
                options = {"shared": [True, False]}
                requires = "liba/0.1@danimtb/testing"

                def configure(self):
                    self.options["*"].shared = self.options.shared
                    self.output.info("shared=%s" % self.options.shared)
                """)

        for without_configure_line in [True, False]:
            if without_configure_line:
                conanfile = conanfile_libb.replace("self.options[", "#")
            else:
                conanfile = conanfile_libb
            client.save({"conanfile.py": conanfile})

            # Test info
            client.run("graph info . -o *:shared=True")
            assert "conanfile.py: shared=True" in client.out
            assert "liba/0.1@danimtb/testing: shared=True" in client.out
            # Test create
            client.run("create . --name=libb --version=0.1 --user=danimtb --channel=testing -o *:shared=True")
            assert "libb/0.1@danimtb/testing: shared=True" in client.out
            assert "liba/0.1@danimtb/testing: shared=True" in client.out
            # Test install
            client.run("install . -o *:shared=True")
            assert "conanfile.py: shared=True" in client.out
            assert "liba/0.1@danimtb/testing: shared=True" in client.out

    def test_define_nested_option_not_freeze(self):
        c = TestClient()
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            class Pkg(ConanFile):
                options = {"without_stacktrace": [True, False],
                           "with_stacktrace_backtrace": [True, False]}
                default_options = {"without_stacktrace": True}
                def configure(self):
                    if self.options.without_stacktrace:
                        del self.options.with_stacktrace_backtrace
                    else:
                        self.options.with_stacktrace_backtrace = True

                def build(self):
                    s = self.options.without_stacktrace
                    self.output.info("without_stacktrace: {}".format(s))

                    if "with_stacktrace_backtrace" in self.options:
                        ss = self.options.get_safe("with_stacktrace_backtrace")
                        self.output.info("with_stacktrace_backtrace: {}".format(ss))
                    else:
                        self.output.info("with_stacktrace_backtrace success deleted!")
            """)
        c.save({"conanfile.py": conanfile})
        c.run("create . --name=pkg --version=0.1")
        assert "pkg/0.1: without_stacktrace: True" in c.out
        assert "pkg/0.1: with_stacktrace_backtrace success deleted!" in c.out
        c.run("create . --name=pkg --version=0.1 -o pkg*:without_stacktrace=False")
        assert "pkg/0.1: without_stacktrace: False" in c.out
        assert "pkg/0.1: with_stacktrace_backtrace: True" in c.out

    def test_del_options_configure(self):
        """
        this test was failing because Options was protecting against removal of options with
        already assigned values. This has been relaxed, to make possible this case
        """
        c = TestClient()
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            class Pkg(ConanFile):
                options = {
                    "shared": [True, False],
                    "fPIC": [True, False],
                }
                default_options = {
                    "shared": False,
                    "fPIC": True,
                }
                def configure(self):
                    if self.options.shared:
                        del self.options.fPIC
            """)
        c.save({"conanfile.py": conanfile})
        c.run("create . --name=pkg --version=0.1")
        c.save({"conanfile.py": GenConanfile("consumer", "1.0").with_requirement("pkg/0.1")},
               clean_first=True)
        c.run("install . -o pkg*:shared=True --build=missing")
        assert "pkg/0.1" in c.out  # Real test is the above doesn't crash

    def test_any(self):
        c = TestClient()
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            class EqualerrorConan(ConanFile):
                name = "equal"
                version = "1.0.0"
                options = {"opt": "ANY"}
                default_options = {"opt": "b=c"}

                def generate(self):
                    self.output.warning("OPTION %s" % self.options.opt)
            """)
        c.save({"conanfile.py": conanfile})
        c.run("install .", assert_error=True)
        assert "Error while initializing options. 'b=c' is not a valid 'options.opt' value." in c.out
        assert "Possible values are ['A', 'N', 'Y']" in c.out


class TestOptionsPriorities:
    # https://github.com/conan-io/conan/issues/11571

    @pytest.fixture
    def _client(self):
        lib1 = textwrap.dedent("""
            from conan import ConanFile

            class Lib1Conan(ConanFile):
                name = "lib1"
                version = "1.0"
                options = {"foobar": [True, False]}
                default_options = {"foobar": False}
            """)
        lib2 = textwrap.dedent("""
            from conan import ConanFile

            class Lib2Conan(ConanFile):
                name = "lib2"
                version = "1.0"
                options = {"logic_for_foobar": [True, False]}
                default_options = {"logic_for_foobar": False}

                def requirements(self):
                    self.requires("lib1/1.0")

                def configure(self):
                    self.options["lib1/*"].foobar = self.options.logic_for_foobar
            """)
        c = TestClient()

        c.save({"lib1/conanfile.py": lib1,
                "lib2/conanfile.py": lib2})

        c.run("create lib1 -o lib1/*:foobar=True")
        c.run("create lib1 -o lib1/*:foobar=False")
        c.run("create lib2 -o lib2/*:logic_for_foobar=True")
        c.run("create lib2 -o lib2/*:logic_for_foobar=False")
        return c

    @staticmethod
    def _app(lib1, lib2, configure):
        app = textwrap.dedent("""
           from conan import ConanFile

           class App(ConanFile):

              def requirements(self):
                  self.requires("{}/1.0")
                  self.requires("{}/1.0")

              def {}(self):
                  self.options["lib2/*"].logic_for_foobar = True
                  self.options["lib1/*"].foobar = False

              def generate(self):
                  self.output.info("LIB1 FOOBAR: {{}}".format(
                                                 self.dependencies["lib1"].options.foobar))
                  self.output.info("LIB2 LOGIC: {{}}".format(
                                              self.dependencies["lib2"].options.logic_for_foobar))
           """)
        return app.format(lib1, lib2, configure)

    def test_profile_priority(self, _client):
        c = _client
        c.save({"app/conanfile.py": self._app("lib1", "lib2", "not_configure")})
        # This order works, because lib1 is expanded first, it takes foobar=False
        c.run("install app -o lib2*:logic_for_foobar=True -o lib1*:foobar=False")
        assert "conanfile.py: LIB1 FOOBAR: False" in c.out
        assert "conanfile.py: LIB2 LOGIC: True" in c.out

        # Now swap order
        c.save({"app/conanfile.py": self._app("lib2", "lib1", "not_configure")})
        c.run("install app -o lib2*:logic_for_foobar=True -o lib1*:foobar=False")
        assert "conanfile.py: LIB1 FOOBAR: False" in c.out
        assert "conanfile.py: LIB2 LOGIC: True" in c.out

    def test_lib1_priority(self, _client):
        c = _client
        c.save({"app/conanfile.py": self._app("lib1", "lib2", "not_configure")})
        # This order works, because lib1 is expanded first, it takes foobar=False
        c.run("install app")
        assert "conanfile.py: LIB1 FOOBAR: False" in c.out
        assert "conanfile.py: LIB2 LOGIC: False" in c.out
        c.run("install app -o lib1*:foobar=True")
        assert "conanfile.py: LIB1 FOOBAR: True" in c.out
        assert "conanfile.py: LIB2 LOGIC: False" in c.out
        c.run("install app -o lib2*:logic_for_foobar=True")
        assert "conanfile.py: LIB1 FOOBAR: False" in c.out
        assert "conanfile.py: LIB2 LOGIC: True" in c.out

    def test_lib2_priority(self, _client):
        c = _client
        c.save({"app/conanfile.py": self._app("lib2", "lib1", "not_configure")})
        # This order works, because lib1 is expanded first, it takes foobar=False
        c.run("install app")
        assert "conanfile.py: LIB1 FOOBAR: False" in c.out
        assert "conanfile.py: LIB2 LOGIC: False" in c.out
        c.run("install app -o lib1*:foobar=True")
        assert "conanfile.py: LIB1 FOOBAR: True" in c.out
        assert "conanfile.py: LIB2 LOGIC: False" in c.out
        c.run("install app -o lib2*:logic_for_foobar=True")
        assert "conanfile.py: LIB1 FOOBAR: True" in c.out
        assert "conanfile.py: LIB2 LOGIC: True" in c.out

    def test_consumer_configure_priority(self, _client):
        c = _client
        c.save({"app/conanfile.py": self._app("lib1", "lib2", "configure")})
        c.run("install app")
        assert "conanfile.py: LIB1 FOOBAR: False" in c.out
        assert "conanfile.py: LIB2 LOGIC: True" in c.out

        # Now swap order
        c.save({"app/conanfile.py": self._app("lib1", "lib2", "configure")})
        c.run("install app")
        assert "conanfile.py: LIB1 FOOBAR: False" in c.out
        assert "conanfile.py: LIB2 LOGIC: True" in c.out


def test_configurable_default_options():
    # https://github.com/conan-io/conan/issues/11487
    c = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        class Pkg(ConanFile):
            name = "pkg"
            version = "0.1"
            settings = "os"
            options = {"backend": [1, 2, 3]}
            def config_options(self):
                if self.settings.os == "Windows":
                    self.options.backend = 2
                else:
                    self.options.backend = 3
            def package_info(self):
                self.output.info("Package with option:{}!".format(self.options.backend))
        """)
    c.save({"conanfile.py": conanfile})
    c.run("create . -s os=Windows")
    assert "pkg/0.1: Package with option:2!" in c.out
    c.run("create . -s os=Windows -o pkg*:backend=3")
    assert "pkg/0.1: Package with option:3!" in c.out
    c.run("create . -s os=Linux")
    assert "pkg/0.1: Package with option:3!" in c.out
    c.run("create . -s os=Windows -o pkg*:backend=1")
    assert "pkg/0.1: Package with option:1!" in c.out

    consumer = textwrap.dedent("""
        from conan import ConanFile
        class Pkg(ConanFile):
            name = "consumer"
            version = "0.1"
            requires = "pkg/0.1"
            def configure(self):
                self.options["pkg*"].backend = 1
        """)
    c.save({"conanfile.py": consumer})
    c.run("install . -s os=Windows")
    assert "pkg/0.1: Package with option:1!" in c.out
    c.run("create . -s os=Windows")
    assert "pkg/0.1: Package with option:1!" in c.out

    # This fails in Conan 1.X
    c.run("create . -s os=Windows -o pkg*:backend=3")
    assert "pkg/0.1: Package with option:3!" in c.out


class TestMultipleOptionsPatterns:
    """
    https://github.com/conan-io/conan/issues/13240
    """

    dep = textwrap.dedent("""
        from conan import ConanFile
        class Pkg(ConanFile):
           version = "1.0"
           options = {"shared": [True, False]}
           default_options = {"shared": False}
           def package_info(self):
               self.output.info(f"SHARED: {self.options.shared}!!")
        """)

    def test_multiple_options_patterns_cli(self):
        """
        https://github.com/conan-io/conan/issues/13240
        """
        c = TestClient()
        consumer = textwrap.dedent("""
            from conan import ConanFile
            class Pkg(ConanFile):
              settings = "os", "arch", "compiler", "build_type"

              def requirements(self):
                  self.requires("dep1/1.0")
                  self.requires("dep2/1.0")
                  self.requires("dep3/1.0")
                  self.requires("dep4/1.0")
            """)
        c.save({"dep/conanfile.py": self.dep,
                "pkg/conanfile.py": consumer})
        for d in (1, 2, 3, 4):
            c.run(f"export dep --name dep{d}")

        # match in order left to right
        c.run('install pkg -o *:shared=True -o dep1*:shared=False -o dep2*:shared=False -b missing')
        assert "dep1/1.0: SHARED: False!!" in c.out
        assert "dep2/1.0: SHARED: False!!" in c.out
        assert "dep3/1.0: SHARED: True!!" in c.out
        assert "dep4/1.0: SHARED: True!!" in c.out

        # All match in order, left to right
        c.run('install pkg -o dep1*:shared=False -o dep2*:shared=False -o *:shared=True -b missing')
        assert "dep1/1.0: SHARED: True!!" in c.out
        assert "dep2/1.0: SHARED: True!!" in c.out
        assert "dep3/1.0: SHARED: True!!" in c.out
        assert "dep4/1.0: SHARED: True!!" in c.out

    def test_multiple_options_patterns(self):
        c = TestClient()
        configure_consumer = textwrap.dedent("""
            from conan import ConanFile
            class Pkg(ConanFile):
                settings = "os", "arch", "compiler", "build_type"

                def requirements(self):
                    self.requires("dep1/1.0")
                    self.requires("dep2/1.0")
                    self.requires("dep3/1.0")
                    self.requires("dep4/1.0")

                def configure(self):
                    self.options["*"].shared = True
                    # Without * also works, equivalent to dep1/*
                    self.options["dep1"].shared = False
                    self.options["dep2*"].shared = False
            """)
        c.save({"dep/conanfile.py": self.dep,
                "pkg/conanfile.py": configure_consumer})
        for d in (1, 2, 3, 4):
            c.run(f"export dep --name dep{d}")

        c.run("install pkg --build=missing")
        assert "dep1/1.0: SHARED: False!!" in c.out
        assert "dep2/1.0: SHARED: False!!" in c.out
        assert "dep3/1.0: SHARED: True!!" in c.out
        assert "dep4/1.0: SHARED: True!!" in c.out

    def test_multiple_options_patterns_order(self):
        c = TestClient()
        configure_consumer = textwrap.dedent("""
            from conan import ConanFile
            class Pkg(ConanFile):
                settings = "os", "arch", "compiler", "build_type"

                def requirements(self):
                    self.requires("dep1/1.0")
                    self.requires("dep2/1.0")
                    self.requires("dep3/1.0")
                    self.requires("dep4/1.0")

                def configure(self):
                    self.options["dep1*"].shared = False
                    self.options["dep2*"].shared = False
                    self.options["*"].shared = True
            """)
        c.save({"dep/conanfile.py": self.dep,
                "pkg/conanfile.py": configure_consumer})
        for d in (1, 2, 3, 4):
            c.run(f"export dep --name dep{d}")

        c.run("install pkg --build=missing")
        assert "dep1/1.0: SHARED: True!!" in c.out
        assert "dep2/1.0: SHARED: True!!" in c.out
        assert "dep3/1.0: SHARED: True!!" in c.out
        assert "dep4/1.0: SHARED: True!!" in c.out

    def test_pattern_version_range_warn(self):
        c = TestClient(light=True)
        c.save({"conanfile.py": GenConanfile("foo", "1.0")})
        c.run("create -o=foo/[>1]:shared=True")
        assert "WARN: risk: Pattern foo/[>1] contains a version range" in c.out

    def test_pattern_version_range_wrong_split(self):
        c = TestClient(light=True)
        c.save({"conanfile.py": GenConanfile("foo", "1.0")})
        c.run("create -o=foo/[>=1]:shared=True", assert_error=True)
        # We split by the first =, so this version range we can't catch and warn before it breaks
        assert "option 'foo/[>' doesn't exist" in c.out


class TestTransitiveOptionsShared:
    """
    https://github.com/conan-io/conan/issues/13854
    """
    @pytest.fixture()
    def client(self):
        c = TestClient()
        c.save({"toollib/conanfile.py": GenConanfile("toollib", "0.1").with_shared_option(False),
                "tool/conanfile.py": GenConanfile("tool", "0.1").with_shared_option(False)
                                                                .with_requires("toollib/0.1"),
                "dep2/conanfile.py": GenConanfile("dep2", "0.1").with_shared_option(False)
                                                                .with_tool_requires("tool/0.1"),
                "dep1/conanfile.py": GenConanfile("dep1", "0.1").with_shared_option(False)
                                                                .with_requires("dep2/0.1"),
                "pkg/conanfile.py": GenConanfile("pkg", "0.1").with_shared_option(False)
                                                              .with_requires("dep1/0.1"),
                "app/conanfile.txt": "[requires]\npkg/0.1"})
        c.run("export toollib")
        c.run("export tool")
        c.run("export dep2")
        c.run("export dep1")
        c.run("export pkg")
        return c

    @staticmethod
    def check(client):
        # tools and libs in build context do not get propagated the options
        for dep in ("toollib", "tool"):
            client.run(f"list {dep}/*:*")
            assert "shared: False" in client.out
        # But the whole host context does
        for dep in ("dep1", "dep2", "pkg"):
            client.run(f"list {dep}/*:*")
            assert "shared: True" in client.out

    def test_transitive_options_shared_cli(self, client):
        client.run("install app --build=missing -o *:shared=True")
        self.check(client)

    def test_transitive_options_shared_profile(self, client):
        client.save({"profile": "[options]\n*:shared=True"})
        client.run("install app --build=missing -pr=profile")
        self.check(client)

    def test_transitive_options_conanfile_txt(self, client):
        client.save({"app/conanfile.txt": "[requires]\npkg/0.1\n[options]\n*:shared=True\n"})
        client.run("install app --build=missing")
        self.check(client)

    def test_transitive_options_conanfile_py(self, client):
        client.save({"app/conanfile.py": GenConanfile().with_requires("pkg/0.1")
                                                       .with_default_option("*:shared", True)})
        client.run("install app/conanfile.py --build=missing")
        self.check(client)

    def test_transitive_options_conanfile_py_create(self, client):
        conanfile = GenConanfile("app", "0.1").with_requires("pkg/0.1") \
                                              .with_default_option("*:shared", True)
        client.save({"app/conanfile.py": conanfile})
        client.run("create app --build=missing")
        self.check(client)


def test_options_no_user_channel_patterns():
    c = TestClient()
    conanfile = textwrap.dedent("""\
        from conan import ConanFile
        class Pkg(ConanFile):
            options = {"myoption": [1, 2, 3]}
            def configure(self):
                self.output.info(f"MYOPTION: {self.options.myoption}")
            """)
    c.save({"dep/conanfile.py": conanfile,
            "pkg/conanfile.py": GenConanfile("pkg", "1.0").with_requires("dep1/0.1", "dep2/0.1@user",
                                                                         "dep3/0.1@user/channel")})
    c.run("export dep --name=dep1 --version=0.1")
    c.run("export dep --name=dep2 --version=0.1 --user=user")
    c.run("export dep --name=dep3 --version=0.1 --user=user --channel=channel")

    c.run("graph info pkg -o *:myoption=3 -o *@:myoption=1")
    assert "dep1/0.1: MYOPTION: 1" in c.out
    assert "dep2/0.1@user: MYOPTION: 3" in c.out
    assert "dep3/0.1@user/channel: MYOPTION: 3" in c.out

    # Recall that order is also important latest matching pattern wins
    c.run("graph info pkg -o *@:myoption=1 -o *:myoption=1")
    assert "dep1/0.1: MYOPTION: 1" in c.out
    assert "dep2/0.1@user: MYOPTION: 1" in c.out
    assert "dep3/0.1@user/channel: MYOPTION: 1" in c.out

    # This is a bit weird negation approach, but it works = all packages that have user channel
    c.run("graph info pkg -o *:myoption=3 -o ~*@:myoption=1")
    assert "dep1/0.1: MYOPTION: 3" in c.out
    assert "dep2/0.1@user: MYOPTION: 1" in c.out
    assert "dep3/0.1@user/channel: MYOPTION: 1" in c.out

    # Which is identical to '~*@' == '*@*'
    c.run("graph info pkg -o *:myoption=3 -o *@*:myoption=1")
    assert "dep1/0.1: MYOPTION: 3" in c.out
    assert "dep2/0.1@user: MYOPTION: 1" in c.out
    assert "dep3/0.1@user/channel: MYOPTION: 1" in c.out


def test_package_options_negate_patterns():
    c = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        class Dep(ConanFile):
            version = "0.1"
            options = {"myoption": [1, 2, 3]}
            def build(self):
                self.output.info(f"MYOPTION={self.options.myoption}!!!")
            """)
    c.save({"dep/conanfile.py": conanfile,
            "pkg/conanfile.py": GenConanfile().with_requires("dep1/0.1", "dep2/0.1", "dep3/0.1")})
    c.run("export dep --name=dep1")
    c.run("export dep --name=dep2")
    c.run("export dep --name=dep3")

    c.run("install pkg --build=* -o *:myoption=1 -o ~dep1/*:myoption=2")
    assert "dep1/0.1: MYOPTION=1!!!" in c.out
    assert "dep2/0.1: MYOPTION=2!!!" in c.out
    assert "dep3/0.1: MYOPTION=2!!!" in c.out

    # Order does matter for options with *:myoption patter, evaluates last
    c.run("install pkg --build=* -o ~dep1/*:myoption=2 -o *:myoption=1")
    assert "dep1/0.1: MYOPTION=1!!!" in c.out
    assert "dep2/0.1: MYOPTION=1!!!" in c.out
    assert "dep3/0.1: MYOPTION=1!!!" in c.out

    # dep3 comes later, works
    c.run("install pkg --build=* -o *:myoption=1 -o ~dep1/*:myoption=2 -o dep3/*:myoption=3")
    assert "dep1/0.1: MYOPTION=1!!!" in c.out
    assert "dep2/0.1: MYOPTION=2!!!" in c.out
    assert "dep3/0.1: MYOPTION=3!!!" in c.out

    # dep3 comes first, then last !dep1 pattern prevails
    c.run("install pkg --build=* -o *:myoption=1 -o dep3/*:myoption=3 -o ~dep1/*:myoption=2")
    assert "dep1/0.1: MYOPTION=1!!!" in c.out
    assert "dep2/0.1: MYOPTION=2!!!" in c.out
    assert "dep3/0.1: MYOPTION=2!!!" in c.out


class TestTransitiveOptionsSharedInvisible:
    """
    https://github.com/conan-io/conan/issues/13854
    When a requirement is visible=False
    """
    @pytest.fixture()
    def client(self):
        c = TestClient()
        c.save({"dep2/conanfile.py": GenConanfile("dep2", "0.1").with_shared_option(False),
                "dep1/conanfile.py": GenConanfile("dep1", "0.1").with_shared_option(False)
                                                                .with_requirement("dep2/0.1",
                                                                                  visible=False),
                "pkg/conanfile.py": GenConanfile("pkg", "0.1").with_shared_option(False)
                                                              .with_requires("dep1/0.1"),
                "app/conanfile.txt": "[requires]\npkg/0.1"})
        c.run("export dep2")
        c.run("export dep1")
        c.run("export pkg")
        return c

    @staticmethod
    def check(client, value):
        for dep in ("dep1", "pkg"):
            client.run(f"list {dep}/*:*")
            assert f"shared: True" in client.out

        # dep2 cannot be affected from downstream conanfile consumers options, only from profile
        client.run(f"list dep2/*:*")
        assert f"shared: {value}" in client.out

    def test_transitive_options_shared_cli(self, client):
        client.run("install app --build=missing -o *:shared=True")
        self.check(client, True)

    def test_transitive_options_shared_profile(self, client):
        client.save({"profile": "[options]\n*:shared=True"})
        client.run("install app --build=missing -pr=profile")
        self.check(client, True)

    def test_transitive_options_conanfile_txt(self, client):
        client.save({"app/conanfile.txt": "[requires]\npkg/0.1\n[options]\n*:shared=True\n"})
        client.run("install app --build=missing")
        self.check(client, False)

    def test_transitive_options_conanfile_py(self, client):
        client.save({"app/conanfile.py": GenConanfile().with_requires("pkg/0.1")
                                                       .with_default_option("*:shared", True)})
        client.run("install app/conanfile.py --build=missing")
        self.check(client, False)

    def test_transitive_options_conanfile_py_create(self, client):
        conanfile = GenConanfile("app", "0.1").with_requires("pkg/0.1") \
                                              .with_default_option("*:shared", True)
        client.save({"app/conanfile.py": conanfile})
        client.run("create app --build=missing")
        self.check(client, False)


class TestImportantOptions:
    @pytest.mark.parametrize("pkg", ["liba", "libb", "app"])
    def test_important_options(self, pkg):
        c = TestClient()

        liba = GenConanfile("liba", "0.1").with_option("myoption", [1, 2, 3])
        libb = GenConanfile("libb", "0.1").with_requires("liba/0.1")
        app = GenConanfile().with_requires("libb/0.1")
        if pkg == "liba":
            liba.with_default_option("myoption!", 2)
        elif pkg == "libb":
            libb.with_default_option("*:myoption!", 2)
        elif pkg == "app":
            app.with_default_option("*:myoption!", 2)
        package_id = textwrap.dedent("""
            def package_id(self):
                self.output.info(f"MYOPTION: {self.info.options.myoption}")
            """)
        liba = str(liba) + textwrap.indent(package_id, "    ")

        c.save({"liba/conanfile.py": liba,
                "libb/conanfile.py": libb,
                "app/conanfile.py": app})
        c.run("export liba")
        c.run("export libb")

        c.run("graph info app -o *:myoption=3")
        assert "liba/0.1: MYOPTION: 2" in c.out

        # And the profile can always important-override the option
        c.run("graph info app -o *:myoption!=3")
        assert "liba/0.1: MYOPTION: 3" in c.out

    def test_profile_shows_important(self):
        c = TestClient()
        c.run("profile show  -o *:myoption!=3")
        assert "*:myoption!=3" in c.out

    def test_important_options_recipe_priority(self):
        c = TestClient()

        liba = GenConanfile("liba", "0.1").with_option("myoption", [1, 2, 3, 4])\
                                          .with_default_option("myoption!", 1)
        libb = GenConanfile("libb", "0.1").with_requires("liba/0.1")\
                                          .with_default_option("*:myoption!", 2)
        app = GenConanfile().with_requires("libb/0.1").with_default_option("*:myoption!", 3)

        package_id = textwrap.dedent("""
            def package_id(self):
                self.output.info(f"MYOPTION: {self.info.options.myoption}")
            """)
        liba = str(liba) + textwrap.indent(package_id, "    ")

        c.save({"liba/conanfile.py": liba,
                "libb/conanfile.py": libb,
                "app/conanfile.py": app})
        c.run("export liba")
        c.run("export libb")

        c.run("graph info app")
        assert "liba/0.1: MYOPTION: 3" in c.out

        c.run("graph info app -o *:myoption!=4")
        assert "liba/0.1: MYOPTION: 4" in c.out

    def test_important_options_recipe_priority_conditional(self):
        c = TestClient()

        liba = textwrap.dedent("""
            from conan import ConanFile
            class Lib(ConanFile):
                name = "liba"
                version = "0.1"
                settings = "os"
                options = {"myoption": [1, 2, 3]}

                def config_options(self):
                    if self.settings.os == "Windows":
                        setattr(self.options, "myoption!", 2)

                def package_id(self):
                    self.output.info(f"MYOPTION: {self.info.options.myoption}")
            """)

        c.save({"liba/conanfile.py": liba,
                "app/conanfile.py": GenConanfile().with_requires("liba/0.1")})
        c.run("export liba")

        c.run("graph info app -s os=Linux -o *:myoption=3")
        assert "liba/0.1: MYOPTION: 3" in c.out
        c.run("graph info app -s os=Linux -o *:myoption=2")
        assert "liba/0.1: MYOPTION: 2" in c.out
        c.run("graph info app -s os=Linux -o *:myoption=1")
        assert "liba/0.1: MYOPTION: 1" in c.out

        c.run("graph info app -s os=Windows -o *:myoption=3")
        assert "liba/0.1: MYOPTION: 2" in c.out
        c.run("graph info app -s os=Windows -o *:myoption=2")
        assert "liba/0.1: MYOPTION: 2" in c.out
        c.run("graph info app -s os=Windows -o *:myoption=1")
        assert "liba/0.1: MYOPTION: 2" in c.out

    def test_wrong_option_syntax_no_trace(self):
        tc = TestClient(light=True)
        tc.save({"conanfile.py": GenConanfile().with_option("myoption", [1, 2, 3])})
        tc.run('create . -o="&:myoption"', assert_error=True)
        assert "ValueError" not in tc.out
        assert "Error while parsing option" in tc.out


class TestConflictOptionsWarnings:

    @pytest.mark.parametrize("important", [True, False])
    def test_options_warnings(self, important):
        c = TestClient()
        liba = GenConanfile("liba", "0.1").with_option("myoption", [1, 2, 3], default=1)
        libb = GenConanfile("libb", "0.1").with_requires("liba/0.1")
        if important:
            libc = GenConanfile("libc", "0.1").with_requirement("liba/0.1") \
                                              .with_default_option("liba*:myoption!", 2)
        else:
            libc = GenConanfile("libc", "0.1").with_requirement("liba/0.1", options={"myoption": 2})
        app = GenConanfile().with_requires("libb/0.1", "libc/0.1")

        c.save({"liba/conanfile.py": liba,
                "libb/conanfile.py": libb,
                "libc/conanfile.py": libc,
                "app/conanfile.py": app})
        c.run("export liba")
        c.run("export libb")
        c.run("export libc")

        c.run("graph info app")
        expected = textwrap.dedent("""\
            Options conflicts
                liba/0.1:myoption=1 (current value)
                    libc/0.1->myoption=2
                It is recommended to define options values in profiles, not in recipes
            """)
        assert expected in c.out
        assert "WARN: risk: There are options conflicts in the dependency graph" in c.out


def test_get_safe_none_option_checks():
    tc = TestClient(light=True)
    tc.save({"conanfile.py": GenConanfile("foo", "1.0")
            .with_package("self.output.info(f'get_safe is None: {self.options.get_safe(\"myoption\") is None}')",
                          "self.output.info(f'get_safe is not None: {self.options.get_safe(\"myoption\") is not None}')",
                          "self.output.info(f'get_safe == None: {self.options.get_safe(\"myoption\") == None}')",
                          "self.output.info(f'get_safe != None: {self.options.get_safe(\"myoption\") != None}')" )})
    tc.run("create .")
    assert "get_safe is None: True" in tc.out
    assert "get_safe is not None: False" in tc.out
    assert "get_safe == None: True" in tc.out
    assert "get_safe != None: False" in tc.out


def test_option_apply_version_range():
    c = TestClient(light=True)
    c.save({"conanfile.py": GenConanfile("dep", "0.1").with_shared_option(False)})
    c.run("create -o shared=True")
    c.run("install --requires=dep/0.1 -o shared=True")  # This worked without problem
    c.run("install --requires=dep/[*] -o shared=True")
    assert "WARN: risk" not in c.out
    # This failed because of dep/[*] not matching pattern, now it works
    assert "Install finished successfully" in c.out



import textwrap

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient


def test_basic():
    client = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        class TestConan(ConanFile):
            name = "hello"
            version = "1.2"
            user = "myuser"
            channel = "mychannel"
        """)
    client.save({"conanfile.py": conanfile})
    client.run("export .")
    assert "hello/1.2@myuser/mychannel" in client.out
    client.run("list *")
    assert "hello/1.2@myuser/mychannel" in client.out
    client.run("create .")
    assert "hello/1.2@myuser/mychannel" in client.out
    assert "hello/1.2:" not in client.out



def test_export_cci():
    c = TestClient(light=True)
    c.save({"conanfile.py": GenConanfile("mypkg", "cci.123")})
    c.run("export .")
    assert ("WARN: risk: Version 'cci.123' contains an alphanumeric "
            "major alongside a minor version") in c.out

    c.save({"conanfile.py": GenConanfile("mypkg", "develop")})
    c.run("export .")
    assert "WARN" not in c.out

    c.save({"conanfile.py": GenConanfile("mypkg", "cci.123")
           .with_class_attribute("package_id_embed_mode='full_mode'")})
    c.run("export .")
    assert "WARN" not in c.out



import pytest
from collections import OrderedDict
from time import sleep

from conan.api.model import RecipeReference
from conan.internal.paths import CONANFILE
from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient, TestServer


class TestExportsSourcesMissing:

    def test_exports_sources_missing(self):
        client = TestClient(default_server_user=True)
        client.save({"conanfile.py": GenConanfile().with_exports_sources("*"),
                     "source.txt": "somesource"})
        client.run("create . --name=pkg --version=0.1 --user=user --channel=testing")
        client.run("upload pkg/0.1@user/testing -r default")

        # Failure because remote is removed
        servers = OrderedDict(client.servers)
        servers["new_server"] = TestServer(users={"user": "password"})
        client2 = TestClient(servers=servers, inputs=["user", "password"])
        client2.run("install --requires=pkg/0.1@user/testing")
        client2.run("remote remove default")
        client2.run("upload pkg/0.1@user/testing -r=new_server", assert_error=True)
        assert "The 'pkg/0.1@user/testing' package has 'exports_sources' but sources " \
                      "not found in local cache." in client2.out
        assert "Probably it was installed from a remote that is no longer available." \
                      in client2.out

        # Failure because remote removed the package
        client2 = TestClient(servers=servers, inputs=2*["admin", "password"])
        client2.run("install --requires=pkg/0.1@user/testing")
        client2.run("remove * -r=default -c")
        client2.run("upload pkg/0.1@user/testing -r=new_server", assert_error=True)
        assert "pkg/0.1@user/testing Error while compressing: The 'pkg/0.1@user/testing' " \
                      in client2.out
        assert "The 'pkg/0.1@user/testing' package has 'exports_sources' but sources " \
                      in client2.out
        assert "Probably it was installed from a remote that is no longer available." \
                      in client2.out


class TestMultiRemotes:

    @pytest.fixture(autouse=True)
    def setup(self):
        self.servers = OrderedDict()
        self.servers["default"] = TestServer()
        self.servers["local"] = TestServer()

    @staticmethod
    def _create(client, number, version, modifier=""):
        files = {CONANFILE: str(GenConanfile(number, version)) + modifier}
        client.save(files, clean_first=True)
        client.run("export . --user=lasote --channel=stable")

    def test_conan_install_build_flag(self):
        """
        Checks conan install --update works with different remotes
        """
        client_a = TestClient(servers=self.servers, inputs=2*["admin", "password"])
        client_b = TestClient(servers=self.servers, inputs=2*["admin", "password"])

        # Upload hello0 to local and default from client_a
        self._create(client_a, "hello0", "0.0")
        client_a.run("upload hello0/0.0@lasote/stable -r local --only-recipe")
        client_a.run("upload hello0/0.0@lasote/stable -r default --only-recipe")
        sleep(1)  # For timestamp and updates checks

        # Download hello0 from local with client_b
        client_b.run("install --requires=hello0/0.0@lasote/stable -r local --build missing")

        # Update hello0 with client_a and reupload
        self._create(client_a, "hello0", "0.0", modifier="\n")
        client_a.run("upload hello0/0.0@lasote/stable -r local --only-recipe")
        assert "Uploading recipe 'hello0/0.0@lasote/stable" in client_a.out

        # Execute info method in client_b, should advise that there is an update
        client_b.run("graph info --requires=hello0/0.0@lasote/stable --check-updates")
        assert "recipe: Update available" in client_b.out
        assert "binary: Cache" in client_b.out

        # Now try to update the package with install -u
        client_b.run("install --requires=hello0/0.0@lasote/stable -u --build='*'")
        assert "hello0/0.0@lasote/stable#64fd8ae21db9eff69c6c681b0e2fc178 - Updated" \
                      in client_b.out

        # Upload a new version from client A, but only to the default server (not the ref-listed)
        # Upload hello0 to local and default from client_a
        sleep(1)  # For timestamp and updates checks
        self._create(client_a, "hello0", "0.0", modifier="\n\n")
        client_a.run("upload hello0/0.0@lasote/stable#latest -r default --only-recipe")

        # Now client_b checks for updates without -r parameter
        # TODO: cache2.0 conan info not yet implemented with new cache
        client_b.run("graph info --requires=hello0/0.0@lasote/stable --check-updates")
        assert "recipe: Update available" in client_b.out
        # assert "Recipe: Cache" in client_b.out

        # But if we connect to default, should tell us that there is an update IN DEFAULT!
        # TODO: cache2.0 conan info not yet implemented with new cache
        client_b.run("graph info --requires=hello0/0.0@lasote/stable -r default --check-updates")
        # assert "Remote: local" in client_b.out
        assert "recipe: Update available" in client_b.out

        # Well, now try to update the package with -r default -u
        client_b.run("install --requires=hello0/0.0@lasote/stable -r default -u --build='*'")
        assert "hello0/0.0@lasote/stable: Forced build from source" \
                      in str(client_b.out)
        # TODO: cache2.0 conan info not yet implemented with new cache
        client_b.run("graph info --requires=hello0/0.0@lasote/stable -u")
        assert "recipe: Cache" in client_b.out
        assert "binary: Cache" in client_b.out

    def test_conan_install_update(self):
        """
        Checks conan install --update works only with the remote associated
        """
        client = TestClient(servers=self.servers, inputs=2*["admin", "password"])

        self._create(client, "hello0", "0.0")
        default_remote_rev = client.exported_recipe_revision()
        client.run("install --requires=hello0/0.0@lasote/stable --build missing")

        client.run("upload hello0/0.0@lasote/stable -r default")
        sleep(1)  # For timestamp and updates checks
        self._create(client, "hello0", "0.0", modifier=" ")
        local_remote_rev = client.exported_recipe_revision()
        client.run("install --requires=hello0/0.0@lasote/stable --build missing")

        client.run("upload hello0/0.0@lasote/stable#latest -r local")
        client.run("remove '*' -c")

        client.run("install --requires=hello0/0.0@lasote/stable")
        # If we don't set a remote we find between all remotes and get the first match
        assert f"hello0/0.0@lasote/stable#{default_remote_rev} - Downloaded" in client.out
        client.run("install --requires=hello0/0.0@lasote/stable --update")
        assert f"hello0/0.0@lasote/stable#{local_remote_rev} - Updated" in client.out

        client.run("install --requires=hello0/0.0@lasote/stable --update -r default")
        assert f"hello0/0.0@lasote/stable#{local_remote_rev} - Newer" \
                      in client.out

        sleep(1)  # For timestamp and updates checks
        # Check that it really updates in case of newer package uploaded to the associated remote
        client_b = TestClient(servers=self.servers, inputs=3*["admin", "password"])
        self._create(client_b, "hello0", "0.0", modifier="  ")
        new_local_remote_rev = client_b.exported_recipe_revision()
        client_b.run("install --requires=hello0/0.0@lasote/stable --build missing")
        client_b.run("upload hello0/0.0@lasote/stable -r local")
        client.run("install --requires=hello0/0.0@lasote/stable --update")
        assert f"hello0/0.0@lasote/stable#{new_local_remote_rev} - Updated" in client.out


class TestMultiRemote:

    @pytest.fixture(autouse=True)
    def setup(self):
        self.servers = OrderedDict()
        self.users = {}
        for i in range(3):
            test_server = TestServer()
            self.servers["remote%d" % i] = test_server
            self.users["remote%d" % i] = [("admin", "password")]

        self.client = TestClient(servers=self.servers, inputs=3*["admin", "password"])

    def test_fail_when_not_notfound(self):
        """
        If a remote fails with a 404 it has to keep looking in the next remote, but if it fails by
        any other reason it has to stop
        """
        servers = OrderedDict()
        servers["s0"] = TestServer()
        servers["s1"] = TestServer()
        servers["s2"] = TestServer()

        client = TestClient(servers=servers)
        client.save({"conanfile.py": GenConanfile("mylib", "0.1")})
        client.run("create . --user=lasote --channel=testing")
        client.run("remote login s1 admin -p password")
        client.run("upload mylib* -r s1 -c")

        servers["s1"].fake_url = "http://asdlhaljksdhlajkshdljakhsd.com"  # Do not exist
        client2 = TestClient(servers=servers)
        client2.run("install --requires=mylib/0.1@conan/testing --build=missing", assert_error=True)
        assert "mylib/0.1@conan/testing: Checking remote: s0" in client2.out
        assert "mylib/0.1@conan/testing: Checking remote: s1" in client2.out
        assert "Unable to connect to remote s1=http://asdlhaljksdhlajkshdljakhsd.com" \
                      in client2.out
        # s2 is not even tried
        assert "mylib/0.1@conan/testing: Trying with 's2'..." not in client2.out

    def test_install_from_remotes(self):
        for i in range(3):
            ref = RecipeReference.loads("hello%d/0.1@lasote/stable" % i)
            self.client.save({"conanfile.py": GenConanfile("hello%d" % i, "0.1")})
            self.client.run("export . --user=lasote --channel=stable")
            self.client.run("upload %s -r=remote%d" % (str(ref), i))

        # Now install it in other machine from remote 0
        client2 = TestClient(servers=self.servers)

        refs = ["hello0/0.1@lasote/stable", "hello1/0.1@lasote/stable", "hello2/0.1@lasote/stable"]
        client2.save({"conanfile.py": GenConanfile("helloX", "0.1").with_requires(*refs)})
        client2.run("install . --build=missing")
        client2.assert_listed_require({"hello0/0.1@lasote/stable": "Downloaded (remote0)",
                                       "hello1/0.1@lasote/stable": "Downloaded (remote1)",
                                       "hello2/0.1@lasote/stable": "Downloaded (remote2)",
                                       })



import platform
import textwrap
import os

import pytest

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient


@pytest.mark.skipif(platform.system() not in ["Windows"], reason="Requires Windows")
@pytest.mark.parametrize("scope", ["build", "run", None])
def test_vcvars_generator(scope):
    client = TestClient(path_with_spaces=False)

    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.microsoft import VCVars

        class TestConan(ConanFile):
            settings = "os", "compiler", "arch", "build_type"

            def generate(self):
                VCVars(self).generate({})
    """.format('scope="{}"'.format(scope) if scope else ""))

    client.save({"conanfile.py": conanfile})
    client.run('install . -s os=Windows -s compiler="msvc" -s compiler.version=191 '
               '-s compiler.cppstd=14 -s compiler.runtime=static')

    assert os.path.exists(os.path.join(client.current_folder, "conanvcvars.bat"))
    assert r"VC\Auxiliary\Build\vcvarsall.bat" in client.load("conanvcvars.bat")
    bat_contents = client.load("conanbuild.bat")
    if scope in ("build", None):
        assert "conanvcvars.bat" in bat_contents


@pytest.mark.skipif(platform.system() not in ["Windows"], reason="Requires Windows")
def test_vcvars_generator_skip():
    """
    tools.microsoft.msbuild:installation_path=disabled avoids creation of conanvcvars.bat
    """
    client = TestClient()
    client.save({"conanfile.py": GenConanfile().with_generator("VCVars")
                                               .with_settings("os", "compiler",
                                                              "arch", "build_type"),
                 "profile": 'include(default)\n[conf]\ntools.microsoft.msbuild:installation_path='})

    client.run('install . -c tools.microsoft.msbuild:installation_path=""')
    assert not os.path.exists(os.path.join(client.current_folder, "conanvcvars.bat"))
    client.run('install . -pr=profile')
    assert not os.path.exists(os.path.join(client.current_folder, "conanvcvars.bat"))


@pytest.mark.skipif(platform.system() not in ["Linux"], reason="Requires Linux")
def test_vcvars_generator_skip_on_linux():
    """
    Skip creation of conanvcvars.bat on Linux build systems
    """
    client = TestClient()
    client.save({"conanfile.txt": "[generators]\nVCVars"})

    client.run('install . -s os=Windows -s compiler=msvc -s compiler.version=193 '
               '-s compiler.runtime=dynamic')
    assert not os.path.exists(os.path.join(client.current_folder, "conanvcvars.bat"))


@pytest.mark.skipif(platform.system() not in ["Windows"], reason="Requires Windows")
def test_vcvars_generator_string():
    client = TestClient(path_with_spaces=False)
    client.save({"conanfile.txt": "[generators]\nVCVars"})
    client.run('install . -s os=Windows -s compiler="msvc" -s compiler.version=191 '
               '-s compiler.cppstd=14 -s compiler.runtime=static')

    assert os.path.exists(os.path.join(client.current_folder, "conanvcvars.bat"))


@pytest.mark.skipif(platform.system() != "Windows", reason="Requires Windows")
@pytest.mark.xfail(reason="we dont have vs2015 installed in gh actions")
def test_vcvars_2015_error():
    # https://github.com/conan-io/conan/issues/9888
    client = TestClient(path_with_spaces=False)
    client.save({"conanfile.txt": "[generators]\nVCVars"})
    client.run('install . -s os=Windows -s compiler="msvc" -s compiler.version=190 '
               '-s compiler.cppstd=14 -s compiler.runtime=static')

    vcvars = client.load("conanvcvars.bat")
    assert 'vcvarsall.bat"  amd64' in vcvars
    assert "-vcvars_ver" not in vcvars


@pytest.mark.skipif(platform.system() != "Windows", reason="Requires Windows")
def test_vcvars_platform_x86():
    # https://github.com/conan-io/conan/issues/11144
    client = TestClient(path_with_spaces=False)
    client.save({"conanfile.txt": "[generators]\nVCVars"})
    client.run('install . -s os=Windows -s compiler="msvc" -s compiler.version=193 '
               '-s compiler.cppstd=14 -s compiler.runtime=static -s:b arch=x86')

    vcvars = client.load("conanvcvars.bat")
    assert 'vcvarsall.bat"  x86_amd64' in vcvars


@pytest.mark.skipif(platform.system() != "Windows", reason="Requires Windows")
def test_vcvars_winsdk_version():
    client = TestClient(path_with_spaces=False)
    client.save({"conanfile.txt": "[generators]\nVCVars"})
    client.run('install . -s os=Windows -s compiler=msvc -s compiler.version=193 '
               '-s compiler.cppstd=14 -s compiler.runtime=static '
               '-c tools.microsoft:winsdk_version=10.0')

    vcvars = client.load("conanvcvars.bat")
    assert 'vcvarsall.bat"  amd64 10.0 -vcvars_ver=14.3' in vcvars


@pytest.mark.skipif(platform.system() != "Windows", reason="Requires Windows")
def test_vcvars_compiler_update():
    client = TestClient(path_with_spaces=False)
    client.save({"conanfile.txt": "[generators]\nVCVars"})
    client.run('install . -s os=Windows -s compiler=msvc -s compiler.version=193 '
               '-s compiler.cppstd=14 -s compiler.runtime=static '
               '-s compiler.update=3')

    vcvars = client.load("conanvcvars.bat")
    assert 'vcvarsall.bat"  amd64 -vcvars_ver=14.33' in vcvars


@pytest.mark.skipif(platform.system() != "Windows", reason="Requires Windows")
def test_vcvars_armv8_windows_store():
    client = TestClient(path_with_spaces=False)
    client.save({"conanfile.txt": "[generators]\nVCVars"})
    client.run('install . -s:b os=Windows -s compiler="msvc" -s compiler.version=194 '
               '-s compiler.cppstd=14 -s compiler.runtime=static -s:h arch=armv8 '
               '-s:h os=WindowsStore -s:h os.version=10.0')

    vcvars = client.load("conanvcvars.bat")
    assert 'vcvarsall.bat"  amd64_arm64' in vcvars


@pytest.mark.skipif(platform.system() != "Windows", reason="Requires Windows")
def test_vcvars_clang_visual2026():
    client = TestClient(path_with_spaces=False)
    client.save({"conanfile.txt": "[generators]\nVCVars"})
    client.run('install . -s:b os=Windows -s compiler=clang -s compiler.version=20 '
               '-s compiler.cppstd=14 -s compiler.runtime=static -s arch=x86_64 '
               '-s compiler.runtime_version=v145 '
               # Using a known existing path to avoid auto-detection via vswhere
               '-c tools.microsoft.msbuild:installation_path=C:/')

    vcvars = client.load("conanvcvars.bat")
    assert '-vcvars_ver=14.5' in vcvars



import textwrap

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient


def test_components_cycles():
    """c -> b -> a -> c"""
    c = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile

        class TestcycleConan(ConanFile):
            name = "testcycle"
            version = "1.0"

            def package_info(self):
                self.cpp_info.components["c"].requires = ["b"]
                self.cpp_info.components["b"].requires = ["a"]
                self.cpp_info.components["a"].requires = ["c"] # cycle!
        """)
    test_conanfile = GenConanfile().with_test("pass").with_generator("CMakeDeps")\
        .with_settings("build_type")
    c.save({"conanfile.py": conanfile,
            "test_package/conanfile.py": test_conanfile})
    c.run("create .", assert_error=True)
    out = c.out
    assert "ERROR: Error in generator 'CMakeDeps': error generating context for 'testcycle/1.0': " \
           "There is a dependency loop in 'self.cpp_info.components' requires:" in out
    assert "a requires c" in out
    assert "b requires a" in out
    assert "c requires b" in out


def test_components_cycle_complex():
    """
    Cycle: a -> b -> c -> d -> b
    Isolated j declaring its libs
    """
    c = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile

        class TestcycleConan(ConanFile):
            name = "testcycle"
            version = "1.0"

            def package_info(self):
                self.cpp_info.components["a"].requires = ["b"]
                self.cpp_info.components["b"].requires = ["c"]
                self.cpp_info.components["c"].requires = ["d"]
                self.cpp_info.components["d"].requires = ["b"]  # cycle!
                self.cpp_info.components["j"].libs = ["libj"]
        """)
    test_conanfile = GenConanfile().with_test("pass").with_generator("CMakeDeps") \
        .with_settings("build_type")
    c.save({"conanfile.py": conanfile,
            "test_package/conanfile.py": test_conanfile})
    c.run("create .", assert_error=True)
    out = c.out
    assert "ERROR: Error in generator 'CMakeDeps': error generating context for 'testcycle/1.0': " \
           "There is a dependency loop in 'self.cpp_info.components' requires:" in out
    assert "a requires b" in out
    assert "b requires c" in out
    assert "c requires d" in out
    assert "d requires b" in out


def test_components_cycles_error():
    c = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile

        class TestcycleConan(ConanFile):
            name = "testcycle"
            version = "1.0"

            def package_info(self):
                self.cpp_info.components["c"].requires = ["b", "d"]
                self.cpp_info.components["b"].requires = ["a"]
                self.cpp_info.components["a"].requires = ["c"] # cycle!
                self.cpp_info.components["d"].includedirs = []
        """)
    test_conanfile = GenConanfile().with_test("pass").with_generator("CMakeDeps")\
        .with_settings("build_type")
    c.save({"conanfile.py": conanfile,
            "test_package/conanfile.py": test_conanfile})
    c.run("create .", assert_error=True)
    assert "ERROR: Error in generator 'CMakeDeps': error generating context for 'testcycle/1.0': " \
           "There is a dependency loop in 'self.cpp_info.components' requires:" in c.out
    assert "a requires c" in c.out
    assert "b requires a" in c.out
    assert "c requires b" in c.out


def test_components_not_required():
    """
    Allow requiring and building against one component, but not propagating it
    https://github.com/conan-io/conan/issues/12965

    """
    c = TestClient()
    conanfile = textwrap.dedent("""
           from conan import ConanFile

           class TestcycleConan(ConanFile):
            name = "wayland"
            version = "1.0"
            requires = "expat/1.0"

            def package_info(self):
                self.cpp_info.components["wayland-scanner"].libdirs = []
        """)
    c.save({"expat/conanfile.py": GenConanfile("expat", "1.0"),
            "wayland/conanfile.py": conanfile})
    c.run("create expat")
    c.run("create wayland")
    assert "wayland/1.0: Created package" in c.out


def test_components_overrides():
    """
    overrides are not direct dependencies, and as such, they don't need to be mandatory
    to specify in the components requires

    https://github.com/conan-io/conan/issues/13922
    """
    c = TestClient()
    consumer = textwrap.dedent("""
        from conan import ConanFile

        class ConanRecipe(ConanFile):
            name = "app"
            version = "0.1"

            def requirements(self):
                self.requires("libffi/3.4.4", override=True)
                self.requires("glib/2.76.2")

            def package_info(self):
                self.cpp_info.requires = ["glib::glib"]
        """)
    c.save({"libffi/conanfile.py": GenConanfile("libffi"),
           "glib/conanfile.py": GenConanfile("glib", "2.76.2").with_requires("libffi/3.0"),
            "app/conanfile.py": consumer})
    c.run("create libffi --version=3.0")
    c.run("create libffi --version=3.4.4")
    c.run("create glib")
    # This used to crash, because override was not correctly excluded
    c.run("create app")
    assert "app/0.1: Created package" in c.out

def test_duplication_component_properties():
    """ Regression for PR 17503 - component lists would be incorrectly aggregated """
    tc = TestClient(light=True)

    dep = textwrap.dedent("""
    from conan import ConanFile

    class Dep(ConanFile):
        name = "dep"
        version = "0.1"

        def package_info(self):
            self.cpp_info.components["acomp"].set_property("prop_list", ["value1"])
            self.cpp_info.components["bcomp"].set_property("prop_list", ["value2"])
            self.cpp_info.components["ccomp"].set_property("prop_list", ["value3"])
    """)

    conanfile = textwrap.dedent("""
    from conan import ConanFile

    class Pkg(ConanFile):
        name = "pkg"
        version = "0.1"
        requires = "dep/0.1"

        def generate(self):
            # Calling this would break property lists of the last lex sorted component
            aggregated_components = self.dependencies["dep"].cpp_info.aggregated_components()
            ccomp = self.dependencies["dep"].cpp_info.components["ccomp"]
            self.output.info("ccomp list: " + str(ccomp.get_property("prop_list")))
    """)

    tc.save({"dep/conanfile.py": dep,
             "conanfile.py": conanfile})
    tc.run("create dep")
    tc.run("create .")
    # The bug would give ccomp the prop_list values of the other two components
    assert "pkg/0.1: ccomp list: ['value3', 'value2', 'value1']" not in tc.out
    assert "pkg/0.1: ccomp list: ['value3']" in tc.out



import os
import shutil

import pytest

from conan.test.utils.tools import TestClient


class TestEncodings:

    @pytest.mark.parametrize("filename", ["conanfile_utf8.txt",
                                          "conanfile_utf8_with_bom.txt",
                                          "conanfile_utf16le_with_bom.txt",
                                          "conanfile_utf16be_with_bom.txt"])
    def test_encoding(self, filename):
        c = TestClient()
        path = os.path.join(os.path.dirname(__file__), "files", filename)
        shutil.copy(path, os.path.join(c.current_folder, "conanfile.txt"))
        c.run("install .")
        assert "Installing packages" in c.out

    def test_error(self):
        c = TestClient()
        conanfile = b"\x81\x8D\x8F\x90\x9D"
        open(os.path.join(c.current_folder, "conanfile.txt"), "wb").write(conanfile)
        c.run("install .", assert_error=True)
        assert "ERROR: Cannot load conanfile.txt" in c.out
        assert "It is recommended to use utf-8 encoding" in c.out


class TestProfileEncodings:

    def test_encoding(self):
        c = TestClient()
        c.save({"conanfile.txt": ""})
        # BOM for utf-7
        open(os.path.join(c.current_folder, "profile"), "wb").write(b'\x2b\x2f\x76\x38')
        c.run("install . -pr=profile")
        assert "Installing packages" in c.out

    def test_error(self):
        c = TestClient()
        c.save({"conanfile.txt": ""})
        open(os.path.join(c.current_folder, "profile"), "wb").write(b"\x81\x8D\x8F\x90\x9D")
        c.run("install . -pr=profile", assert_error=True)
        assert "ERROR: Cannot load profile" in c.out
        assert "It is recommended to use utf-8 encoding" in c.out
